/* nsdperf - Test network performance, simulating GPFS NSD client/server

   See the README file for information about building and running
   this program.
*/

#ifndef _GNU_SOURCE
#define _GNU_SOURCE
#endif
#define _LINUX_SOURCE_COMPAT

#include <algorithm>
#include <cstdio>
#include <cstdlib>
#include <deque>
#include <fstream>
#include <iomanip>
#include <iostream>
#include <list>
#include <map>
#include <new>
#include <queue>
#include <set>
#include <sstream>
#include <string>
#include <vector>

#include <math.h>
#include <netdb.h>
#include <pthread.h>
#include <signal.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <sys/uio.h>
#include <sys/errno.h>
#include <sys/ioctl.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/un.h>
#ifdef _AIX
#include <libperfstat.h>
#endif

#ifdef AF_INET6
#define IPV6_SUPPORT
#endif

#ifdef __sparc
#define strerror_r(_e,_b,_s) strerror(_e)
#define INADDR_NONE INADDR_BROADCAST
#include <sys/filio.h>
#endif

using namespace std;


// Sized types
typedef unsigned char          UChar;
typedef short                  Int16;
typedef int                    Int32;
typedef long long              Int64;
typedef unsigned short         UInt16;
typedef unsigned int           UInt32;
typedef unsigned long long     UInt64;

// High-resolution time (microsecond units)
typedef long long HTime;

// These definitions allow compatibility with the Windows Socket Library.
// They work in conjunction with a GPFS interface library.
#ifndef USE_WINSOCK
  // Socket type
  typedef int Sock;

  // Socket event description for the poll() function
  typedef struct pollfd PollSock;

  // The customary invalid socket value.
  static const Sock INVALID_SOCK = -1;
#endif

// Message ID type
typedef UInt32 MsgId;

// Program version
static const string version = "$Revision: 1.11 $";

// Default port to use 
static const int NSDPERF_PORT = 6668;

// Default number of message worker threads
static const int MSG_WORKERS = 10;

// Default number of tester threads
static const unsigned int TESTER_THREADS = 4;

// Maximum size for data buffer
static const unsigned int MAX_BUFFSIZE = 8 * 1024 * 1024;

// Maximum TCP send/receive buffer size
static const int MAX_SOCKSIZE = 100 * 1024 * 1024;

// Maximum number of FDs to pass to poll
static const int MAX_POLLFD_NUM = 8192;

// Size of a message header
static const unsigned int MSG_HDRSIZE = 4 * sizeof(UInt32);

// Magic number for message headers
static const UInt32 MSG_MAGIC = 0x1F2E3D4CU;

// Maximum possible value for an HTime
static const long long MAX_HTIME = 0x7FFFFFFFFFFFFFFFLL;

// For ioctl arguments
static const int on = 1;

// Flag to use in sendmsg system call to prevent EPIPE errors.  Some
// systems don't have this.
#ifdef MSG_NOSIGNAL
static const int SENDMSG_FLAGS = MSG_NOSIGNAL;
#else
static const int SENDMSG_FLAGS = 0;
#endif


// Forward declarations
class MsgRecord;
struct RcvMsg;
class Receiver;
struct ReplyEntry;
struct Target;
struct TestReq;
class Tester;
class Thread;


// Error macros
#define Warn(msg) do \
  { cerr << progname << ": Warning: " << msg << endl; }\
  while (false)
#define Warnm(msg) do \
  { cerr << progname << ": Warning: " << msg << ": "; perror(""); }\
  while (false)
#define Error(msg) do \
  { cerr << progname << ": " << msg << endl; exit(EXIT_FAILURE); }\
  while (false)
#define Errorm(msg) do \
  { cerr << progname << ": " << msg << ": "; perror(""); exit(EXIT_FAILURE); }\
  while (false)

// Console log output macros
#define Log(msg) do \
  { cout << msg << endl; }\
  while (false)
#define Logt(msg) do \
  { cout << getTime() << " " << msg << endl; }\
  while (false)
#define Logm(msg) do \
  { int e = errno; cout << msg << ": " << geterr(e) << endl; }\
  while (false)


// Error codes
enum Errno
{
  E_OK,
  E_INVAL,
  E_NOENT,
  E_WOULDBLOCK,
  E_CONNRESET,
  E_BADMSG,
  E_BROKEN,
  E_SENDFAILED,
  E_REPLY
};


// Address types for IpAddr
enum Atype
{
  AT_IPV4,
  AT_IPV6
};

  
// Structure for holding an IPv4 or IPv6 address
struct IpAddr
{
  Atype fam;
  UChar a[16];
  IpAddr();
  bool operator<(const IpAddr &iaddr) const;
  bool operator!=(const IpAddr &iaddr) const;
  Errno parse(const string hostname);
  void loadSockaddr(const sockaddr *saddrP);
  string tostring() const;
#ifdef IPV6_SUPPORT
  int getFamily() const { return (fam == AT_IPV6) ? AF_INET6 : AF_INET; }
#else
  int getFamily() const { return AF_INET; }
#endif
};


// Abstract base class for threads.  Derived class must supply the routine
// body.
class Thread
{
  bool running;
  pthread_t th;

public:
  virtual ~Thread() {}
  void init();
  void startup();
  pthread_t getThread() const { return th; }
  virtual int threadBody() = 0;
};


// Data buffer that allows putting and getting items for sending in messages
class DataBuff
{
  char *buffP;                  // The data buffer
  unsigned int alloc;           // Actual allocated length
  unsigned int bufflen;         // Length currently assigned to buffer
  unsigned int buffpos;         // Position for put/get

public:
  DataBuff()  { buffP = NULL; alloc = bufflen = buffpos = 0; }
  DataBuff(unsigned int len)           { alloc = 0; newBuff(len); }
  DataBuff(char *dP, unsigned int len) { alloc = 0; initBuff(dP, len); }
  ~DataBuff() { if (alloc > 0) delete [] buffP; }
  void newBuff(unsigned int len);
  void initBuff(char *dataP, int datalen);
  void resetBuff() { buffpos = 0; }
  char *getBuffP() { return buffP; }
  unsigned int getBufflen() const { return bufflen; }
  void putUInt32(UInt32 i);
  void putUInt64(UInt64 i);
  void putIpAddr(IpAddr iaddr);
  void putString(string s);
  UInt32 getUInt32();
  UInt64 getUInt64();
  IpAddr getIpAddr();
  string getString();
};

  
// Gather histogram of response times
class Histogram
{
  map<UInt32, UInt32> buckets;
  int nEvents;
  HTime totalTime;

public:
  Histogram() { nEvents = 0; totalTime = 0; }
  void addEntry(HTime t);
  void addHist(const Histogram *hP);
  void printHist(ostream &os) const;
  void putBuff(DataBuff *dbP) const;
  void getBuff(DataBuff *dbP);
  unsigned int calcLen() const
    { return sizeof(UInt64) + 2 * sizeof(UInt32) +
        2 * buckets.size() * sizeof(UInt32); }
  double average() const
    { return static_cast<double>(totalTime) / nEvents; }
  UInt32 median() const;
  UInt32 minTime() const
    { return buckets.empty() ? 0 : buckets.begin()->first; }
  UInt32 maxTime() const
    { return buckets.empty() ? 0 : buckets.rbegin()->first; }
  UInt32 maxBucket() const;
};


// Compare two histogram bucket entries by bucket value  
struct BucketCmp
{
  bool operator()(pair<UInt32, UInt32> p1, pair<UInt32, UInt32> p2) const
  { return p1.second < p2.second; }
};

  
// Receive states of a connection
enum RState
{
  rcv_idle,
  rcv_header,
  rcv_data
};
  

// Message type codes
enum MType
{
  mtUnknown,            // Unknown
  mtReply,              // Reply message
  mtReplyErr,           // Error reply message
  mtVersion,            // Query version
  mtEcho,               // Echo data
  mtWrite,              // Accept buffer of test data
  mtRead,               // Send back a buffer of test data
  mtNwrite,             // NSD style write: target asks for test data
  mtKill,               // Tell server to exit
  mtConnect,            // Connect to specified servers
  mtReset,              // Close any existing server connections
  mtParms,              // Set test parameters
  mtTest,               // Run performance test to all servers
  mtStatus,             // Return node status
  mtStatOn,             // Turn on test statistics gathering
  mtStatOff,            // Turn off test statistics gathering
  mtIdlePct,            // Get idle CPU percentage from last test
  mtLast                // Highest message type number
};
  

// An entry in the queue of waiters for exclusive use of a socket for sending
struct InuseWaiter
{
  // Wake this when the socket is available
  pthread_cond_t iwCond;

  // The following fields are used to determine priority between waiters
  MType mt;             // Original msgType (not mtReply or mtReplyErr)
  MsgId msgId;          // Message identifier (earlier messages are lower)
  int datalen;          // Length of the message

  InuseWaiter();        // Default constructor - not defined
  InuseWaiter(MType tmt, MsgId tmsgId, int tdatalen);
};

  
// Function object to compare two InuseWaiter entries to determine which
// has higher priority
struct InuseCmp
{ bool operator()(const InuseWaiter *w1P, const InuseWaiter *w2P) const; };


// TCP connection
class TcpConn
{
  pthread_mutex_t connMutex;
  pthread_cond_t connCond;
  priority_queue<InuseWaiter *, vector<InuseWaiter *>, InuseCmp> waiters;

  int refCount;
  Sock tcSock;
  IpAddr dest;
  bool inuse;           // Serializer for senders
  bool broken;          // True if socket was shut down due to error
  Histogram connHist;   // Histogram of response times from last test

  // Receiver thread state
  RState recvState;
  char *recvP;          // Receive pointer
  int recvlen;          // Length of data received so far
  int recvmax;          // Size of receive buffer
  RcvMsg *rmsgP;        // Message will be received here

public:
  TcpConn();            // Default constructor - not defined
  TcpConn(Sock tsock, IpAddr tdest);
  string destName() const { return dest.tostring(); }
  Histogram *getHistP() { return &connHist; }
  bool isBroken() const { return broken; }
  void holdConn();
  void releaseConn();
  void connShutdown();
  Errno receiverEvent();
  Errno recvMessage();
  Errno sendMessage(MType mt, char *dataP, int datalen, MsgRecord *mrP);
  Errno sendit(MType mt, MType origmt, MsgId msgId, char *dataP,
               int datalen, MsgRecord *mrP);
};

  
// Target node for test
struct Target
{
  string hostname;
  IpAddr iaddr;                 // IP address
  TcpConn *connP;               // Connection (null if not connected)
  bool active;                  // For detecting inactive Targets

  Target();                     // Default constructor - not defined
  Target(const Target &m);      // Copy constructor - not defined
  Target(const string thostname, const IpAddr tiaddr)
  {
    hostname = thostname;
    iaddr = tiaddr;
    connP = NULL;
    active = true;
  }
  ~Target()
  {
    if (connP != NULL)
    {
      connP->connShutdown();
      connP->releaseConn();
    }
  }
  string makeConnection();
  RcvMsg *sendm(MType mt, char *dataP, int datalen);
  string name() const;
};


// For sorting Target objects for round-robin scheduling
struct SortedTarget
{
  int tindex;
  Target *targP;
  SortedTarget(int ndx, Target *tP) : tindex(ndx), targP(tP) {}
  static bool comp(const SortedTarget t1, const SortedTarget t2)
  {
    if (t1.tindex != t2.tindex) return t1.tindex < t2.tindex;
    return t1.targP->iaddr < t2.targP->iaddr;
  }
};

  
// Function object to compare two node table iterators based on host name
// length.
struct TargetNameCmp
{
  bool operator()(pair<const IpAddr, const Target *> p1,
                  pair<const IpAddr, const Target *> p2) const
  { return p1.second->hostname.length() < p2.second->hostname.length(); }
};

  
// Structure to hold received messages
struct RcvMsg
{
  TcpConn *connP;               // The connection that this message came from
  char hdr[MSG_HDRSIZE];        // Buffer for receiving message header
  DataBuff msgBuff;             // Buffer for message data
  MsgId msgId;                  // Message identifier
  MType msgType;                // Message type
  string errText;               // Error message, or empty if no error

  RcvMsg();                     // Default constructor - not defined
  RcvMsg(const RcvMsg &m);      // Copy constructor - not defined
  RcvMsg(TcpConn *tconnP)
  {
    connP = tconnP;
    memset(hdr, 0, MSG_HDRSIZE);
    msgId = 0;
    msgType = mtUnknown;
    connP->holdConn();
  }
  ~RcvMsg()                     { connP->releaseConn(); }
  char *msgBuffP()              { return msgBuff.getBuffP(); }
  unsigned int msgLen() const   { return msgBuff.getBufflen(); }
  void sendReply(DataBuff *dbP, string errText = "");
  bool showError();
  bool startAdminReq();
  void endAdminReq();
  void dispatch();

  // Message handlers
  void handleVersion();
  void handleNwrite();
  void handleConnect();
  void handleReset();
  void handleParms();
  void handleTest();
  void handleStatus();
  void handleIdlePct();
};


// This object is used to wait for pending replies
class MsgRecord
{
  pthread_mutex_t waitMutex;
  pthread_cond_t waitCond;
  set<MsgId> waitTab;             // IDs of pending messages
  list<RcvMsg *> replies;       // Reply data

public:
  MsgRecord();
  ~MsgRecord();
  void addMsg(MsgId msgId);
  void waitForReplies();
  bool checkReplies();
  void gotReply(RcvMsg *rmsgP);
  RcvMsg *nextReply();
};


// This object is saved in the pending reply table to keep track of
// outstanding replies.  Each message has a unique msgId which is used to
// find the ReplyEntry when the reply comes in.
struct ReplyEntry
{
  MsgRecord *mrP;               // MsgRecord to wake up when reply comes in
  TcpConn *connP;               // Connection where the message was sent

  ReplyEntry();                 // Default constructor - not defined
  ReplyEntry(const ReplyEntry &m); // Copy constructor - not defined
  ReplyEntry(MsgRecord *tmrP, TcpConn *tconnP)
  {
    mrP = tmrP;
    connP = tconnP;
    connP->holdConn();
  }
  ~ReplyEntry()
  {
    connP->releaseConn();
  }
};

  
// Test types
enum TType
{
  ttWrite,              // Round-robin write
  ttRead,               // Round-robin read
  ttNwrite,             // NSD-style write
  ttRW,                 // Simultaneous read and write
  ttSwrite,             // Single node write per tester
  ttSread,              // Single node read per tester
  ttLast
};


// Tester thread work request
struct TestReq
{
  TType tt;
  list<Target *> testNodes;
  HTime endTime;
  UInt64 totBytes;
  Histogram hist;
  string errText;

  TestReq(TType ttt, HTime tendTime)
  {
    tt = ttt;
    endTime = tendTime;
    totBytes = 0;
  }
};

  
// A receiver thread
class Receiver : public Thread
{
  pthread_mutex_t receiverMutex;        // Protects socketTab
  pthread_cond_t receiverCond;          // For additions to socketTab
  map<Sock, TcpConn *> socketTab;       // All connections, indexed by socket
  bool fdNeedsRefresh;                  // Signal to rebuild poll array
  Sock rcvSocks[2];                     // Socket pair for waking up poll

public:
  Receiver();
  virtual ~Receiver();
  void addConn(Sock sock, TcpConn *connP);
  void wakeUp();
  virtual int threadBody();
};


// Message worker thread
class MsgWorker : public Thread
{
public:
  MsgWorker() {}
  virtual ~MsgWorker() {}
  virtual int threadBody();
};


// Tester thread
class Tester : public Thread
{
  TestReq *reqP;
  
public:
  Tester() { reqP = NULL; }
  virtual ~Tester() {}
  virtual int threadBody();
  void doTest(TestReq *trP);
};


// Listen/accept thread
class ListenAccept : public Thread
{
  int currsize;                 // Socket buffer size currently in use
public:
  ListenAccept() { currsize = 0; }
  virtual ~ListenAccept() {}
  virtual int threadBody();
  void updateSocksize();
};


// Global variables
static string progname;
static bool littleEndian;
static bool quitflag = false;
static DataBuff testBuff;
static deque<string> commands;
static int nestingLevel = 0;
static int idleTime = -1;
static bool collectStats = false;

static multimap<IpAddr, Target *> serverNodes;
static map<IpAddr, Target *> clientNodes;
static map<IpAddr, Target *> allNodes;

static int debugLevel = 0;
static short port = NSDPERF_PORT;
static int socksize = 0;
static int nWorkers = MSG_WORKERS;
static int nParallel = 1;
static bool server = false;
static bool useipv6 = false;
static int addrFamily;
static int testTime = 10;        
static unsigned int buffsize = 4194304;
static unsigned int nTesterThreads = TESTER_THREADS;
static bool showHist = false;
static string plotFname;

static pthread_mutex_t globalMutex;
static pthread_cond_t globalCond;
static list<Thread *> deadThreads;
static unsigned int nThreadsStarted = 0;
static vector<Receiver *> receiverTab;
static vector<Receiver *>::iterator nextReceiver;
static map<MsgId, ReplyEntry *> pendReplyTab;
static MsgId nextMsgId = 1;
static bool cmdInProgress = false;
static ListenAccept *laThreadP = NULL;

static pthread_mutex_t workerMutex;
static pthread_cond_t workerCond;
static list<RcvMsg *> msgQueue;
static list<RcvMsg *> bulkQueue;
static int workersActive = 0;

static pthread_mutex_t testerMutex;
static pthread_cond_t testerCond;
static set<Tester *> testerTab;
static list<TestReq *> doneList;
static unsigned int nTestersWorking = 0;



// Return current time of day
static HTime getTime()
{
  struct timeval tv;
  gettimeofday(&tv, NULL);
  return tv.tv_sec * 1000000LL + tv.tv_usec;
}


// Swap bytes in a word
static UInt32 ByteSwap32(UInt32 d)
{
  return ((d >> 24) & 0x000000FFU) |
         ((d >>  8) & 0x0000FF00U) |
         ((d <<  8) & 0x00FF0000U) |
         ((d << 24) & 0xFF000000U);
}


// Swap bytes in a double word
static UInt64 ByteSwap64(UInt64 d)
{
  return ((d >> 56) & 0x00000000000000FFULL) |
         ((d >> 40) & 0x000000000000FF00ULL) |
         ((d >> 24) & 0x0000000000FF0000ULL) |
         ((d >>  8) & 0x00000000FF000000ULL) |
         ((d <<  8) & 0x000000FF00000000ULL) |
         ((d << 24) & 0x0000FF0000000000ULL) |
         ((d << 40) & 0x00FF000000000000ULL) |
         ((d << 56) & 0xFF00000000000000ULL);
}


// Report the number of online processors
static int numProcessors()
{
  #ifndef _SC_NPROCESSORS_ONLN
    return 2;
  #else
    return sysconf(_SC_NPROCESSORS_ONLN);
  #endif
}


// Set socket send/receive buffer sizes.  This routine must be called
// before listen or connect.
static void setSockSizes(Sock sock)
{
  if (socksize == 0)
    return;
  if (setsockopt(sock, SOL_SOCKET, SO_RCVBUF, &socksize, sizeof(socksize)) < 0)
    Warnm("setsockopt SO_RCVBUF failed");
  if (setsockopt(sock, SOL_SOCKET, SO_SNDBUF, &socksize, sizeof(socksize)) < 0)
    Warnm("setsockopt SO_SNDBUF failed");
}


// Make a socket non-blocking
static void setSockNonblocking(Sock sock)
{
  int rc = ioctl(sock, (int)FIONBIO, (char *)&on, (int)sizeof(int));
  if (rc < 0)
    Errorm("ioctl FIONBIO");
}


// Get error message for a given error number
static string geterr(int e)
{
  char b[256];
  snprintf(b, sizeof(b), "error %d", e);
  (void)strerror_r(e, b, sizeof(b));
  return b;
}


// Get next word from string starting at pos.  Update pos to end of word.
static string getword(const string s, string::size_type &pos)
{
  string::size_type pos1;

  pos1 = s.find_first_not_of(" \t\r", pos);
  if (pos1 == string::npos)
  {
    pos = pos1;
    return "";
  }
  pos = s.find_first_of(" \t\r", pos1);
  return s.substr(pos1, (pos == string::npos) ? pos : pos - pos1);
}


// Return true if start of line matches specified string
static bool match(const string line, const string s)
{
  return line.length() >= s.length() && line.substr(0, s.length()) == s;
}


// Split input string into separate command lines, delimited by semicolon
// or slash.
static void splitCmd(const string inp, deque<string> *linesP)
{
  string::size_type p1 = 0, p2;
  if (inp.empty())
    return;
  while (true)
  {
    p2 = inp.find_first_of(";/", p1);
    linesP->push_back(inp.substr(p1, (p2 == string::npos) ? p2 : p2 - p1));
    if (p2 == string::npos)
      break;
    p1 = p2 + 1;
  }
}


// Round d to the specified number of significant decimal digits.
// Negative values are not supported.
static double siground(double d, int digits)
{
  int j, n;
  double tpow = 1.0;

  if (d <= 0.0 || digits <= 0) return 0.0;
  for (j = 0; j < digits; j++) tpow *= 10.0;

  // Normalize
  n = 0;
  if (d >= tpow) while (d >= tpow)        d /= 10.0, n--;
  else           while (d * 10.0 < tpow)  d *= 10.0, n++;

  // Round to precision
  d = floor(d + 0.5);

  // Undo normalize
  if (n > 0)      for (j = 0; j < n; j++)  d /= 10.0;
  else if (n < 0) for (j = 0; j < -n; j++) d *= 10.0;

  return d;
}


// Sleep until specified time.  Return current time.
static HTime sleepUntil(HTime endTime)
{
  HTime now;
  for (now = getTime(); now < endTime; now = getTime())
    usleep(endTime - now);
  return now;
}


// Allocate a new data buffer for output.  If a big enough buffer is
// already allocated, just adjust the assigned buffer length.
void DataBuff::newBuff(unsigned int len)
{
  if (len == 0)
    Error("Allocating zero length buffer");

  bufflen = len;
  buffpos = 0;
  if (alloc > 0)
  {
    if (len <= alloc)
      return;
    delete [] buffP;
  }
  buffP = new char[len];
  alloc = len;
}


// Initialize buffer to point to specified data area
void DataBuff::initBuff(char *dataP, int datalen)
{
  if (alloc > 0)
  {
    delete [] buffP;
    alloc = 0;
  }
  buffP = dataP;
  bufflen = datalen;
  buffpos  = 0;
}


// Put a UInt32 item into buffer, performing endian conversion if
// necessary, and bump the buffer pointer.
void DataBuff::putUInt32(UInt32 i)
{
  if (buffpos + sizeof(i) > bufflen) Error("putUInt32 buffer overflow");
  if (littleEndian) i = ByteSwap32(i);
  memcpy(&buffP[buffpos], &i, sizeof(i));
  buffpos += sizeof(i);
}


// Put a UInt64 item into buffer, performing endian conversion if
// necessary, and bump the buffer pointer.
void DataBuff::putUInt64(UInt64 i)
{
  if (buffpos + sizeof(i) > bufflen) Error("putUInt64 buffer overflow");
  if (littleEndian) i = ByteSwap64(i);
  memcpy(&buffP[buffpos], &i, sizeof(i));
  buffpos += sizeof(i);
}


// Put an IpAddr into buffer and bump the buffer pointer.  No endian
// conversion is performed on the address data, since it is kept in memory
// in native byte order.
void DataBuff::putIpAddr(IpAddr iaddr)
{
  if (buffpos + sizeof(UInt32) + sizeof(iaddr.a) > bufflen)
    Error("putIpAddr buffer overflow");
  putUInt32(iaddr.fam);
  memcpy(&buffP[buffpos], iaddr.a, sizeof(iaddr.a));
  buffpos += sizeof(iaddr.a);
}


// Put string into buffer, with length and padding
void DataBuff::putString(string s)
{
  unsigned int slen = s.length();
  unsigned int padlen = 4 - (slen & 3);
  putUInt32(slen + padlen);
  if (buffpos + slen + padlen > bufflen) Error("putString buffer overflow");
  s.copy(&buffP[buffpos], slen);
  for (buffpos += slen; padlen > 0; padlen--)
    buffP[buffpos++] = '\0';
}


// Get UInt32 from buffer
UInt32 DataBuff::getUInt32()
{
  UInt32 i;
  if (buffpos + sizeof(i) > bufflen) Error("getUInt32 buffer underflow");
  memcpy(&i, &buffP[buffpos], sizeof(i));
  buffpos += sizeof(i);
  return littleEndian ? ByteSwap32(i) : i;
}


// Get UInt64 from buffer
UInt64 DataBuff::getUInt64()
{
  UInt64 i;
  if (buffpos + sizeof(i) > bufflen) Error("getUInt64 buffer underflow");
  memcpy(&i, &buffP[buffpos], sizeof(i));
  buffpos += sizeof(i);
  return littleEndian ? ByteSwap64(i) : i;
}


// Get IpAddr from buffer
IpAddr DataBuff::getIpAddr()
{
  IpAddr iaddr;
  if (buffpos + sizeof(UInt32) + sizeof(iaddr.a) > bufflen)
    Error("getIpAddr buffer underflow");
  iaddr.fam = static_cast<Atype>(getUInt32());
  memcpy(iaddr.a, &buffP[buffpos], sizeof(iaddr.a));
  buffpos += sizeof(iaddr.a);
  return iaddr;
}


// Get string from buffer
string DataBuff::getString()
{
  unsigned int slen = getUInt32();
  if (buffpos + slen > bufflen) Error("getString buffer underflow");
  string s(&buffP[buffpos]);
  if (s.length() >= slen) Error("incorrect getString length");
  buffpos += slen;
  return s;
}


// Calculate buffer space needed to do putString
static unsigned int calcLen(string s)
{
  unsigned int slen = s.length();
  unsigned int padlen = 4 - (slen & 3);
  return sizeof(UInt32) + slen + padlen;
}


// Put Histogram into buffer and bump the buffer pointer
void Histogram::putBuff(DataBuff *dbP) const
{
  map<UInt32, UInt32>::const_iterator b;
  dbP->putUInt64(totalTime);
  dbP->putUInt32(nEvents);
  dbP->putUInt32(buckets.size());
  for (b = buckets.begin(); b != buckets.end(); ++b)
  {
    dbP->putUInt32(b->first);
    dbP->putUInt32(b->second);
  }
}


// Get Histogram from buffer
void Histogram::getBuff(DataBuff *dbP)
{
  unsigned int nb;
  UInt32 bnum, count;
  buckets.clear();
  totalTime = dbP->getUInt64();
  nEvents = dbP->getUInt32();
  for (nb = dbP->getUInt32(); nb > 0; nb--)
  {
    bnum = dbP->getUInt32();
    count = dbP->getUInt32();
    buckets[bnum] = count;
  }
}


// Structure to hold CPU utilization statistics in ticks
struct CpuStats
{
  UInt64 user, system, idle, iowait, other;
  UInt64 total() { return user + system + idle + iowait + other; }
};

  
// Get CPU utilization statistics.  Return false if not supported.
static bool getStats(CpuStats *statsP)
{
  static bool supported = true;
  if (!supported)
    return false;

#ifdef _AIX
  perfstat_partition_total_t pstats;
  if (perfstat_partition_total(NULL, &pstats, sizeof(pstats), 1) < 0)
    Errorm("perfstat_partition_total");

  // More work is required to calculate utilization for LPAR in shared mode,
  // so punt for now.
  if (pstats.type.b.shared_enabled)
  {
    supported = false;
    return false;
  }

  statsP->user   = pstats.puser;
  statsP->system = pstats.psys;
  statsP->idle   = pstats.pidle;
  statsP->iowait = pstats.pwait;
  statsP->other  = 0;
  return true;

#else
  ifstream infile;
  string line;
  UInt64 n[8];
  bool gotit = false;

  // Get CPU info from /proc/stat.  See Documentation/filesystems/proc.txt
  // in the Linux kernel source.
  infile.open("/proc/stat");
  if (!infile)
  {
    supported = false;
    return false;
  }
  while (getline(infile, line))
  {
    // Fields are: user nice system idle iowait irq softirq steal
    if (sscanf(line.c_str(), "cpu %llu %llu %llu %llu %llu %llu %llu %llu",
               &n[0], &n[1], &n[2], &n[3], &n[4], &n[5], &n[6], &n[7]) == 8)
    {
      gotit = true;
      break;
    }
    
    // Older kernels didn't have the steal field
    n[7] = 0;
    if (sscanf(line.c_str(), "cpu %llu %llu %llu %llu %llu %llu %llu",
               &n[0], &n[1], &n[2], &n[3], &n[4], &n[5], &n[6]) == 7)
    {
      gotit = true;
      break;
    }
  }
  infile.close();
  if (!gotit)
  {
    supported = false;
    return false;
  }

  statsP->user   = n[0] + n[1];
  statsP->system = n[2];
  statsP->idle   = n[3];
  statsP->iowait = n[4];
  statsP->other  = n[5] + n[6] + n[7];
  return true;
#endif
}


// Return CPU idle percentage since previous call, or -1 if not supported
// on this OS or if this is the first call.
static int cpuIdle()
{
  static bool initialized = false;
  static CpuStats oldstats;
  CpuStats stats;
  UInt64 tot;
  int val = -1;

  if (!getStats(&stats))
    return -1;

  if (initialized)
  {
    tot = stats.total() - oldstats.total();
    if (tot != 0)
      val = (stats.idle - oldstats.idle) * 100 / tot;
  }
  else
    initialized = true;

  oldstats = stats;
  return val;
}


// Convert hostname and IP address to a string for output messages
static string hostString(const string hname, const IpAddr iaddr)
{
  int rc;
  struct sockaddr_storage saddr;

  // Determine whether the host name is a dotted decimal or hex IP address
  // by trying to convert it to a network address structure.  If the
  // conversion is successful, then don't show the IP address in result,
  // since that would be redundant.
  rc = inet_pton(iaddr.getFamily(), hname.c_str(), &saddr);
  if (rc < 0)
    Errorm("inet_pton");
  else if (rc == 0)
    return hname + " (" + iaddr.tostring() + ")";
  return hname;
}


// IpAddr constructor
IpAddr::IpAddr()
{
  in_addr_t addr = INADDR_NONE;
  fam = AT_IPV4;
  memset(a, 0, sizeof(a));
  memcpy(a, &addr, sizeof(addr));
}


// Compare two IP addresses
bool IpAddr::operator<(const IpAddr &iaddr) const
{
  unsigned int j;
  if (fam != iaddr.fam)
    return fam < iaddr.fam;
  for (j = 0; j < sizeof(a); j++)
    if (a[j] != iaddr.a[j])
      return a[j] < iaddr.a[j];
  return false;
}


// Test inequality of two IP addresses
bool IpAddr::operator!=(const IpAddr &iaddr) const
{
  unsigned int j;
  if (fam != iaddr.fam)
    return true;
  for (j = 0; j < sizeof(a); j++)
    if (a[j] != iaddr.a[j])
      return true;
  return false;
}


// Convert a host name string to an IP address.  If an error occurs, log an
// error message and return E_INVAL.
Errno IpAddr::parse(const string hostname)
{
  string::size_type pos = 0;
  string hname = getword(hostname, pos);
#ifdef IPV6_SUPPORT
  struct addrinfo hints, *resP;
  int rc;

  memset(&hints, 0, sizeof(hints));
  hints.ai_family = addrFamily;
  hints.ai_socktype = SOCK_STREAM;

  rc = getaddrinfo(hname.c_str(), NULL, &hints, &resP);
  if (rc != 0)
  {
    Log("Invalid hostname \"" << hname << "\": " << gai_strerror(rc));
    return E_INVAL;
  }

  // Copy the address data.  If multiple addresses are returned, we only
  // use the first one.  It would be better to try all of them until a
  // working address is found.
  loadSockaddr(resP->ai_addr);

  freeaddrinfo(resP);
#else
  unsigned int d[4];
  fam = AT_IPV4;
  memset(a, 0, sizeof(a));
  if (sscanf(hname.c_str(), "%u.%u.%u.%u", &d[0], &d[1], &d[2], &d[3]) == 4)
  {
    if (d[0] > 255 || d[1] > 255 || d[2] > 255 || d[3] > 255)
    {
      Log("Invalid hostname: " << hname);
      return E_INVAL;
    }
    a[0] = d[0]; a[1] = d[1]; a[2] = d[2]; a[3] = d[3];
  }
  else
  {
    struct hostent *hp = gethostbyname(hname.c_str());
    if (hp == NULL || hp->h_addrtype != AF_INET || hp->h_length != 4)
    {
      Log("Cannot find " << hname);
      return E_INVAL;
    }
    memcpy(a, hp->h_addr, 4);
  }
#endif
  return E_OK;
}


// Load IpAddr contents from a sockaddr struct
void IpAddr::loadSockaddr(const sockaddr *saddrP)
{
  switch (saddrP->sa_family)
  {
#ifdef IPV6_SUPPORT
    case AF_INET6:
      fam = AT_IPV6;
      memcpy(a, &reinterpret_cast<const sockaddr_in6 *>(saddrP)->sin6_addr,
             sizeof(a));
      break;
#endif

    case AF_INET:
      fam = AT_IPV4;
      memset(a, 0, sizeof(a));
      memcpy(a, &reinterpret_cast<const sockaddr_in *>(saddrP)->sin_addr,
             sizeof(in_addr_t));
      break;

    default:
      Error("Invalid sa_family");
      break;
  }
}


// Convert IP address to printable string
string IpAddr::tostring() const
{
#ifdef IPV6_SUPPORT
  char buf[INET6_ADDRSTRLEN];
#else
  char buf[INET_ADDRSTRLEN];
#endif
  if (inet_ntop(getFamily(), &a, buf, sizeof(buf)) == NULL)
    Errorm("inet_ntop");
  return buf;  
}


// Error-checking wrappers for mutex and condition variable routines
static void thInitMutex(pthread_mutex_t *mutexP)
{ if (pthread_mutex_init(mutexP, NULL) != 0) Errorm("pthread_mutex_init"); }

static void thInitCond(pthread_cond_t *condP)
{ if (pthread_cond_init(condP, NULL) != 0) Errorm("pthread_cond_init"); }

static void thLock(pthread_mutex_t *mutexP)
{ if (pthread_mutex_lock(mutexP) != 0) Errorm("pthread_mutex_lock"); }

static void thUnlock(pthread_mutex_t *mutexP)
{ if (pthread_mutex_unlock(mutexP) != 0) Errorm("pthread_mutex_unlock"); }

static void thWait(pthread_cond_t *condP, pthread_mutex_t *mutexP)
{ if (pthread_cond_wait(condP, mutexP) != 0) Errorm("pthread_cond_wait"); }

static void thSignal(pthread_cond_t *condP)
{ if (pthread_cond_signal(condP) != 0) Errorm("pthread_cond_signal"); }

static void thBcast(pthread_cond_t *condP)
{ if (pthread_cond_broadcast(condP) != 0) Errorm("pthread_cond_broadcast"); }

static void thJoin(pthread_t th, void **valPP)
{ if (pthread_join(th, valPP) != 0) Errorm("pthread_join"); }


// Wrapper for thread body
extern "C" void *thread_wrapper(void *argP)
{
  int rc;
  Thread *thP = static_cast<Thread *>(argP);
  thP->startup();
  rc = thP->threadBody();

  // Tell the main thread to clean us up
  thLock(&globalMutex);
  deadThreads.push_back(thP);
  thBcast(&globalCond);
  thUnlock(&globalMutex);
  pthread_exit(&rc);
  return NULL;
}


// Start the thread and wait for it to get going (it would be nice to do
// this from the constructor, except that virtual functions always use the
// base class version, so we wouldn't be able to call the thread body
// routine).
void Thread::init()
{
  thLock(&globalMutex);
  running = false;
  nThreadsStarted++;
  if (pthread_create(&th, NULL, thread_wrapper, this) != 0)
    Errorm("pthread_create");
  while (!running)
    thWait(&globalCond, &globalMutex);
  thUnlock(&globalMutex);
}


// Call this routine when a new thread starts to tell the init routine
// that the thread is running.
void Thread::startup()
{
  thLock(&globalMutex);
  running = true;
  thBcast(&globalCond);
  thUnlock(&globalMutex);
}


// Receiver constructor
Receiver::Receiver()
{
  thInitMutex(&receiverMutex);
  thInitCond(&receiverCond);
  fdNeedsRefresh = false;

  // Create a socket pair to be used for waking up receiver thread when it
  // is in poll.
  if (socketpair(AF_UNIX, SOCK_STREAM, 0, rcvSocks) < 0)
    Errorm("socketpair");
  setSockNonblocking(rcvSocks[0]);
  setSockNonblocking(rcvSocks[1]);
}


// Receiver destructor
Receiver::~Receiver()
{
  int j;
  for (j = 0; j < 2; j++)
    if (close(rcvSocks[j]) < 0)
      Errorm("close rcvSocks");
}


// Receiver thread body
int Receiver::threadBody()
{
  int j, numfd, nFDs = 0;
  Sock socket = INVALID_SOCK;
  PollSock fdList[MAX_POLLFD_NUM];
  map<Sock, TcpConn *>::iterator siter;
  TcpConn *connP;

  while (true)
  {
    // Wait for the listen thread to give us a socket
    thLock(&receiverMutex);
    while (socketTab.empty())
      thWait(&receiverCond, &receiverMutex);
    thUnlock(&receiverMutex);

    // Loop, polling the sockets
    while (true)
    {
      // Rebuild poll array if socket table has changed
      thLock(&receiverMutex);
      if (fdNeedsRefresh)
      {
        memset(fdList, 0, sizeof(fdList));
        fdList[0].fd = rcvSocks[0];
        fdList[0].events = POLLIN;
        fdList[0].revents = 0;
        nFDs = 1;
        for (siter = socketTab.begin();
             siter != socketTab.end();
             ++siter, nFDs++)
        {
          if (nFDs >= MAX_POLLFD_NUM)
            Error("more than " << MAX_POLLFD_NUM << " connections per receiver");
          fdList[nFDs].fd = siter->first;
          fdList[nFDs].events = POLLIN;
          fdList[nFDs].revents = 0;
        }
        fdNeedsRefresh = false;
      }
      thUnlock(&receiverMutex);
      if (nFDs == 0)
        break;

      // Wait for data from a socket
      while (true)
      {
        numfd = poll(fdList, nFDs, 5 * 1000);
        if (numfd >= 0)
          break;
        if (errno != EINTR)
          Errorm("poll");
      }

      for (j = 0; j < nFDs && numfd > 0; j++)
      {
        if (fdList[j].revents == 0)
          continue;

        numfd--;
        socket = fdList[j].fd;

        // If this is the socket pair, that means someone sent us a wakeup
        // notification.  Remove all data from the socket.
        if (socket == rcvSocks[0])
        {
          char tmpBuf[16];
          while (recv(socket, tmpBuf, sizeof(tmpBuf), 0) > 0);
          continue;
        }

        thLock(&receiverMutex);
        siter = socketTab.find(socket);
        if (siter == socketTab.end())
          Error("socket " << socket << " not found in socketTab");
        connP = siter->second;
        thUnlock(&receiverMutex);

        // Handle event from socket.  Shut it down if error occurs.
        if (connP->receiverEvent() != E_OK)
        {
          thLock(&receiverMutex);
          socketTab.erase(socket);
          fdNeedsRefresh = true;
          thUnlock(&receiverMutex);

          connP->connShutdown();
          connP->releaseConn();
        }
      }
    }
  }
  return 0;
}


// Add a new TCP connection for this receiver to handle
void Receiver::addConn(Sock sock, TcpConn *connP)
{
  connP->holdConn();
  thLock(&receiverMutex);
  if (socketTab.find(sock) != socketTab.end())
    Error("socket " << sock << " is already in socketTab");
  socketTab[sock] = connP;
  fdNeedsRefresh = true;
  thSignal(&receiverCond);
  thUnlock(&receiverMutex);
  wakeUp();
}


// Wake up receiver thread if it is in poll
void Receiver::wakeUp()
{
  int rc;
  char buf[1] = {'x'};
  while (true)
  {
    rc = send(rcvSocks[1], buf, 1, 0);
    if (rc == 0)
      Error("rcvSocks disconnected");
    if (rc > 0 || errno == EAGAIN || errno == EWOULDBLOCK)
      break;
    if (errno != EINTR)
      Errorm("send to rcvSocks");
  }
}


// Perform round-robin selection of receiver threads
static Receiver *pickReceiver()
{
  Receiver *rcvP;
  thLock(&globalMutex);
  rcvP = *nextReceiver;
  if (++nextReceiver == receiverTab.end())
    nextReceiver = receiverTab.begin();
  thUnlock(&globalMutex);
  return rcvP;
}


// InuseWaiter constructor
InuseWaiter::InuseWaiter(MType tmt, MsgId tmsgId, int tdatalen)
{
  thInitCond(&iwCond);
  mt = tmt;
  msgId = tmsgId;
  datalen = tdatalen;
}


// Assign priority according to message type.  Messages with high-valued
// types will be granted use of the socket for sending before lower ones.
// Inbound messages with priority less than 5 are considered bulk messages,
// and will be processed after all other message types.
static int msgPriority(MType mt)
{
  int j;
  static int mpri[mtLast];
  static bool inited = false;

  if (mt <= 0 || mt >= mtLast)
    Error("Invalid message type");
  if (!inited)
  {
    // Set up message priority table for faster lookups
    for (j = 0; j < mtLast; j++)
      switch (j)
      {
        case mtKill:    mpri[j] = 9; break;
        case mtStatOn:  mpri[j] = 8; break;
        case mtStatOff: mpri[j] = 8; break;
        case mtEcho:    mpri[j] = 1; break;
        case mtWrite:   mpri[j] = 1; break;
        case mtRead:    mpri[j] = 1; break;
        case mtNwrite:  mpri[j] = 1; break;
        default:        mpri[j] = 5; break;
      }
    inited = true;
  }
  return mpri[mt];
}


// Convert message type to string for debug output
static string mtToString(MType mt)
{
  switch (mt)
  {
    case mtUnknown:  return "Unknown";
    case mtReply:    return "Reply";
    case mtReplyErr: return "ReplyErr";
    case mtVersion:  return "Version";
    case mtEcho:     return "Echo";
    case mtWrite:    return "Write";
    case mtRead:     return "Read";
    case mtNwrite:   return "Nwrite";
    case mtKill:     return "Kill";
    case mtConnect:  return "Connect";
    case mtReset:    return "Reset";
    case mtParms:    return "Parms";
    case mtTest:     return "Test";
    case mtStatus:   return "Status";
    case mtStatOn:   return "StatOn";
    case mtStatOff:  return "StatOff";
    case mtIdlePct:  return "IdlePct";
    default:         break;
  }
  return "??";
}

   
// Compare two InuseWaiter entries to determine which has higher priority
bool InuseCmp::operator()(const InuseWaiter *w1P, const InuseWaiter *w2P) const
{
  // High priority messages go first
  int p1 = msgPriority(w1P->mt);
  int p2 = msgPriority(w2P->mt);
  if (p1 < p2) return true;
  if (p1 > p2) return false;

  // Short messages go ahead of long ones, but if they are close in size,
  // consider them to be the same.
  p1 = w1P->datalen / 1000;
  p2 = w2P->datalen / 1000;
  if (p1 < p2) return true;
  if (p1 > p2) return false;

  // Older messages go ahead of later ones
  return w1P->msgId < w2P->msgId;
}


// TcpConn constructor
TcpConn::TcpConn(Sock tsock, IpAddr tdest)
{
  thInitMutex(&connMutex);
  thInitCond(&connCond);
  refCount = 0;
  tcSock = tsock;
  dest = tdest;
  inuse = false;
  broken = false;
  recvState = rcv_idle;
  recvP = NULL;
  recvlen = 0;
  recvmax = 0;
  rmsgP = NULL;
}


// Put a hold on this TcpConn so that it won't be deleted
void TcpConn::holdConn()
{
  thLock(&connMutex);
  refCount++;
  if (debugLevel > 1)
    Log("holdConn " << refCount << ", sock " << tcSock);
  thUnlock(&connMutex);
}


// Release hold on this TcpConn and delete it if count is zero.  This is
// the only place that the socket is closed.
void TcpConn::releaseConn()
{
  thLock(&connMutex);
  refCount--;
  if (debugLevel > 1)
    Log("releaseConn " << refCount << ", sock " << tcSock);
  if (refCount > 0)
    thUnlock(&connMutex);
  else
  {
    if (!broken)
      Error("TcpConn released while still connected");
    thUnlock(&connMutex);
    if (close(tcSock) < 0)
      Errorm("socket close");
    if (debugLevel > 0)
      Log("Closed connection to " << destName());
    delete this;
  }
}


// Shut down the TCP connection.  Everyone who is using the connection
// should wake up due to the socket shutdown and release their hold.  When
// the hold count goes to zero, the socket will be closed and the TcpConn
// deleted.  The broken flag ensures that only one thread calls shutdown
// and prevents new threads from starting to use the connection.
void TcpConn::connShutdown()
{
  thLock(&connMutex);
  if (!broken)
  {
    if (shutdown(tcSock, SHUT_RDWR) < 0 && errno != ENOTCONN)
      Errorm("socket shutdown");
    broken = true;

    // Wake up anyone who was waiting for exclusive use of the socket.
    // Wait for them to notice the broken connection and pop themselves off
    // the waiters list.
    while (!waiters.empty())
    {
      thSignal(&waiters.top()->iwCond);
      thWait(&connCond, &connMutex);
    }
  }
  thUnlock(&connMutex);

  // Remove any queued messages for this connection.  If a worker thread
  // already picked up the RcvMsg, then he should fail quickly since the
  // broken flag is set.
  thLock(&workerMutex);
  list<RcvMsg *>::iterator riter;
  list<RcvMsg *> *qP = &msgQueue;
  riter = qP->begin();
  while (true)
  {
    if (riter == qP->end())
    {
      if (qP != &msgQueue)
        break;
      qP = &bulkQueue;
      riter = qP->begin();
      continue;
    }
    
    RcvMsg *rmsgP = *riter;
    if (rmsgP->connP != this)
      ++riter;
    else
    {
      riter = qP->erase(riter);
      delete rmsgP;
    }
  }
  thUnlock(&workerMutex);

  // Scan the pending reply table and wake up anybody who is waiting for a
  // reply on this connection.
  thLock(&globalMutex);
  map<MsgId, ReplyEntry *>::iterator piter;
  for (piter = pendReplyTab.begin(); piter != pendReplyTab.end(); )
  {
    MsgId msgId = piter->first;
    ReplyEntry *reP = piter->second;
    if (reP->connP != this)
      ++piter;
    else
    {
      pendReplyTab.erase(piter++);

      // Manufacture a dummy RcvMsg with errText set and use that to wake
      // up sender.
      RcvMsg *rmsgP = new RcvMsg(this);
      rmsgP->msgId = msgId;
      rmsgP->errText = "connection broken";
      reP->mrP->gotReply(rmsgP);
      delete reP;
    }
  }
  thUnlock(&globalMutex);
}


// Handle an event from poll
Errno TcpConn::receiverEvent()
{
  Errno err = E_OK;
  DataBuff db;

  if (debugLevel > 1)
    Logt("event on sock " << tcSock);

  // Keep reading until the socket blocks or is shut down due to an error
  while (true)
  {
    if (broken)
    {
      err = E_BROKEN;
      break;
    }

    switch (recvState)
    {
      case rcv_idle:
        rmsgP = new RcvMsg(this);
        recvP = rmsgP->hdr;
        recvmax = MSG_HDRSIZE;
        recvState = rcv_header;
        break;

      case rcv_header:
        err = recvMessage();
        if (err != E_OK)
          break;
        
        // Message header format:
        //    UInt32 magic      Magic number, should be MSG_MAGIC
        //    UInt32 msgId      Message identifier
        //    UInt32 msgType    Message type
        //    UInt32 msgLen     Length of data which follows this header

        db.initBuff(rmsgP->hdr, MSG_HDRSIZE);
        if (db.getUInt32() != MSG_MAGIC)
        {
          Log("Invalid message from " << destName());
          err = E_BADMSG;
          break;
        }
        rmsgP->msgId   = static_cast<MsgId>(db.getUInt32());
        rmsgP->msgType = static_cast<MType>(db.getUInt32());
        recvmax        = db.getUInt32();

        if (recvmax > 16 * 1024 * 1024)
          Error("Invalid msgLen");

        if (recvmax > 0)
        {
          rmsgP->msgBuff.newBuff(recvmax);
          recvP = rmsgP->msgBuffP();
        }
        recvState = rcv_data;
        break;

      case rcv_data:
        err = recvMessage();
        if (err != E_OK)
          break;

        // If this is a reply, wake up the sender.  Otherwise, pass the
        // received message to a worker thread.
        if (rmsgP->msgType == mtReply || rmsgP->msgType == mtReplyErr)
        {
          ReplyEntry *reP;
          map<MsgId, ReplyEntry *>::iterator piter;
          thLock(&globalMutex);
          piter = pendReplyTab.find(rmsgP->msgId);
          if (piter == pendReplyTab.end())
            Error("reply not found in pending reply table");
          reP = piter->second;
          pendReplyTab.erase(rmsgP->msgId);
          thUnlock(&globalMutex);

          // If this is an error reply, pull out the error text
          if (rmsgP->msgType == mtReplyErr)
            rmsgP->errText = rmsgP->msgBuff.getString();

          reP->mrP->gotReply(rmsgP);
          delete reP;
        }
        else
        {
          thLock(&workerMutex);
          if (msgPriority(rmsgP->msgType) < 5)
            bulkQueue.push_back(rmsgP);
          else
            msgQueue.push_back(rmsgP);
          thSignal(&workerCond);
          thUnlock(&workerMutex);
        }

        rmsgP = NULL;
        recvState = rcv_idle;
        break;
        
      default:
        Error("invalid receiver state");
    }
    if (err == E_WOULDBLOCK)
      return E_OK;
    if (err != E_OK)
      break;
  }
  if (err != E_OK)
  {
    // Discard any partially received message
    delete rmsgP;
    rmsgP = NULL;
  }
  return err;
}


// This routine is called from a receiver thread to read pending data.
// Data is read into the buffer pointed to by recvP until recvmax bytes
// have been read.  The recvlen variable is used to keep track of how much
// has been read so far.  Returns E_WOULDBLOCK if more data is needed from
// socket.
Errno TcpConn::recvMessage()
{
  while (recvlen < recvmax)
  {
    int rc = recv(tcSock, recvP, recvmax - recvlen, 0);
    if (rc == 0)
    {
      Log("Connection to " << destName() << " broken");
      return E_CONNRESET;
    }
    if (rc < 0 && errno != EAGAIN && errno != EWOULDBLOCK)
    {
      Logm("Receive from " << destName() << " failed");
      return E_CONNRESET;
    }
    if (rc < 0)
      return E_WOULDBLOCK;

    if (debugLevel > 1)
      Logt("received " << rc << " bytes");
    recvlen += rc;
    recvP += rc;
  }
  recvlen = 0;
  recvmax = 0;
  recvP = NULL;
  return E_OK;
}


// Send a message on this connection.  When a reply is received, it will be
// added to the reply list in mrP.  Caller must call waitForReplies before
// deleting the MsgRecord.
Errno TcpConn::sendMessage(MType mt, char *dataP, int datalen, MsgRecord *mrP)
{
  return sendit(mt, mt, 0, dataP, datalen, mrP);
}


// Internal routine used by sendMessage and sendReply to send a message
Errno TcpConn::sendit(MType mt, MType origmt, MsgId msgId, char *dataP,
                      int datalen, MsgRecord *mrP)
{
  char hdr[MSG_HDRSIZE];
  DataBuff db;
  struct msghdr msg;
  struct iovec iov[2];
  int rc, iovCount;
  unsigned int skipBytes;
  PollSock ps;

  // Wait for exclusive use of the socket for sending
  thLock(&connMutex);
  if (inuse)
  {
    InuseWaiter iw(origmt, msgId, datalen);
    waiters.push(&iw);
    while ((inuse && !broken) || waiters.top() != &iw)
      thWait(&iw.iwCond, &connMutex);
    waiters.pop();
    if (broken)
      thSignal(&connCond);
  }
  if (broken)
  {
    thUnlock(&connMutex);
    return E_BROKEN;
  }
  inuse = true;
  thUnlock(&connMutex);

  if (mt != mtReply && mt != mtReplyErr)
  {
    // Assign a unique message ID for this message
    thLock(&globalMutex);
    msgId = nextMsgId++;

    // Record this message ID in the MsgRecord for caller to wait on.  We
    // must do this before adding an entry in pendReplyTab because a socket
    // break will scan pendReplyTab and attempt to remove the entry for
    // this ID from the MsgRecord.
    mrP->addMsg(msgId);

    // Add an entry to pending reply table, which will be used to wake the
    // MsgRecord when the reply comes in.  It will also be used to locate
    // waiters if the connection breaks.
    if (pendReplyTab.find(msgId) != pendReplyTab.end())
      Error("message " << msgId << " is already in pendReplyTab");
    pendReplyTab[msgId] = new ReplyEntry(mrP, this);
    thUnlock(&globalMutex);
  }

  // Build message header
  db.initBuff(hdr, MSG_HDRSIZE);
  db.putUInt32(MSG_MAGIC);
  db.putUInt32(msgId);
  db.putUInt32(mt);
  db.putUInt32(datalen);

  // Combine header and data into an iovec
  iov[0].iov_base = hdr;
  iov[0].iov_len = MSG_HDRSIZE;
  if (datalen == 0)
    iovCount = 1;
  else
  {
    iov[1].iov_base = dataP;
    iov[1].iov_len = datalen;
    iovCount = 2;
  }

  if (debugLevel > 0)
    Logt("sending msg type " << mtToString(mt) << " ID " << msgId
         << " len " << datalen << " to " << destName());

  memset(&msg, 0, sizeof(msg));
  msg.msg_iov = iov;
  msg.msg_iovlen = iovCount;
  skipBytes = 0;

  // Send the header and data
  while (msg.msg_iovlen > 0)
  {
    if (broken)
      break;

    if (skipBytes > 0)
    {
      while (msg.msg_iovlen > 0 && skipBytes >= msg.msg_iov[0].iov_len)
      {
        skipBytes -= msg.msg_iov[0].iov_len;
        msg.msg_iov++;
        msg.msg_iovlen--;
      }
      if (skipBytes == 0)
        continue;
      msg.msg_iov[0].iov_base =
        static_cast<char *>(msg.msg_iov[0].iov_base) + skipBytes;
      msg.msg_iov[0].iov_len -= skipBytes;
      skipBytes = 0;
      continue;
    }

    rc = sendmsg(tcSock, &msg, SENDMSG_FLAGS);
    if (rc == 0)
    {
      Warn("connection to " << destName() << " broken");
      connShutdown();
      break;
    }
    if (rc < 0 && errno != EAGAIN && errno != EWOULDBLOCK)
    {
      Warnm("send to " << destName() << " failed");
      connShutdown();
      break;
    }
    if (rc > 0)
    {
      skipBytes = rc;
      if (debugLevel > 1)
        Logt("write " << rc << " bytes on sock " << tcSock);
      continue;
    }

    // Socket is blocked on send, so poll and try again
    ps.fd = tcSock;
    ps.events = POLLOUT;
    ps.revents = 0;
    while (true)
    {
      rc = poll(&ps, 1, -1);
      if (rc >= 0)
        break;
      if (errno != EINTR)
        Errorm("poll for send");
    }
    skipBytes = 0;
  }

  // Release exclusive use of the socket
  thLock(&connMutex);
  inuse = false;
  if (!waiters.empty())
    thSignal(&waiters.top()->iwCond);
  thUnlock(&connMutex);
  return broken ? E_BROKEN : E_OK;
}


// Create a TCP connection to target node.  Return error message if failure.
string Target::makeConnection()
{
  string errmsg, ver;
  Sock sock;
  int rc;
  RcvMsg *rmsgP;

  if (connP != NULL)
    return errmsg;

  if ((sock = socket(iaddr.getFamily(), SOCK_STREAM, 0)) == INVALID_SOCK)
    Errorm("socket open");
  setSockSizes(sock);

  // Don't delay sending messages
  if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on)) < 0)
    Errorm("setsockopt TCP_NODELAY");

  // Make the connection
#ifdef IPV6_SUPPORT
  if (iaddr.fam == AT_IPV6)
  {
    struct sockaddr_in6 saddr;
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin6_family = iaddr.getFamily();
    saddr.sin6_port = htons(port);
    memcpy(&saddr.sin6_addr, iaddr.a, sizeof(saddr.sin6_addr));
    rc = connect(sock, reinterpret_cast<sockaddr *>(&saddr), sizeof(saddr));
  }
  else
#endif
  {
    struct sockaddr_in saddr;
    memset(&saddr, 0, sizeof(saddr));
    saddr.sin_family = iaddr.getFamily();
    saddr.sin_port = htons(port);
    memcpy(&saddr.sin_addr, iaddr.a, sizeof(saddr.sin_addr));
    rc = connect(sock, reinterpret_cast<sockaddr *>(&saddr), sizeof(saddr));
  }
  if (rc < 0)
  {
    errmsg = geterr(errno);
    close(sock);
    return errmsg;
  }

  // Set for non-blocking IO
  setSockNonblocking(sock);
  
  connP = new TcpConn(sock, iaddr);
  connP->holdConn();
  pickReceiver()->addConn(sock, connP);
  rmsgP = sendm(mtVersion, NULL, 0);
  if (!rmsgP->errText.empty())
  {
    errmsg = rmsgP->errText;
    delete rmsgP;
    return errmsg;
  }

  // Verify that his version matches ours
  ver = rmsgP->msgBuff.getString();
  delete rmsgP;
  if (ver != version)
    return "Version mismatch.  Remote version is \"" + ver + "\".";

  Log("Connected to " << hostname);
  return errmsg;
}


// Send one message, wait for reply, and return it.  If an error occurs,
// put error message in errText.
RcvMsg *Target::sendm(MType mt, char *dataP, int datalen)
{
  Errno err;
  MsgRecord mr;
  err = connP->sendMessage(mt, dataP, datalen, &mr);
  mr.waitForReplies();
  if (err != E_OK)
  {
    RcvMsg *rmsgP = new RcvMsg(connP);
    rmsgP->errText = "connection to " + name() + " broken";
    return rmsgP;
  }
  return mr.nextReply();
}


// Convert host name and IP address to printable name for error messages
string Target::name() const
{
  return hostString(hostname, iaddr);
}


// MsgRecord constructor
MsgRecord::MsgRecord()
{
  thInitMutex(&waitMutex);
  thInitCond(&waitCond);
}


// MsgRecord destructor
MsgRecord::~MsgRecord()
{
  thLock(&waitMutex);
  if (!waitTab.empty())
    Error("MsgRecord destructor called with messages pending");

  // Delete any replies
  while (!replies.empty())
  {
    delete replies.front();
    replies.pop_front();
  }
  thUnlock(&waitMutex);
}


// Add an entry to set of waiting messages
void MsgRecord::addMsg(MsgId msgId)
{
  thLock(&waitMutex);
  if (!waitTab.insert(msgId).second)
    Error("message " << msgId << " is already in waitTab");
  thUnlock(&waitMutex);
}


// Wait for replies from all messages to be received
void MsgRecord::waitForReplies()
{
  thLock(&waitMutex);
  while (!waitTab.empty())
    thWait(&waitCond, &waitMutex);
  thUnlock(&waitMutex);
}


// Wait for any outstanding replies, check all received replies for errors,
// printing a message if found, and then delete the replies.  Return true
// if an error was seen.
bool MsgRecord::checkReplies()
{
  RcvMsg *rmsgP;
  bool gotError = false;
  waitForReplies();
  for (rmsgP = nextReply(); rmsgP != NULL; rmsgP = nextReply())
  {
    if (rmsgP->showError()) gotError = true;
    delete rmsgP;
  }
  return gotError;
}


// Call this when a reply is received
void MsgRecord::gotReply(RcvMsg *rmsgP)
{
  thLock(&waitMutex);
  if (waitTab.erase(rmsgP->msgId) != 1)
    Error("msgId not found in waitTab");
  replies.push_back(rmsgP);
  if (waitTab.empty())
    thSignal(&waitCond);
  thUnlock(&waitMutex);
}


// Return the next reply, and remove it from reply list
RcvMsg *MsgRecord::nextReply()
{
  RcvMsg *rmsgP = NULL;
  thLock(&waitMutex);
  if (!replies.empty())
  {
    rmsgP = replies.front();
    replies.pop_front();
  }
  thUnlock(&waitMutex);
  return rmsgP;
}


// Add a new entry to histogram.  Input should be a response time in HTime
// units (microseconds).  Histogram buckets are currently in units of msec.
void Histogram::addEntry(HTime t)
{
  if (t < 0)
    return;

  nEvents++;
  totalTime += t;

  // Convert time to msec, which is the bucket number
  UInt32 n = (t + 500) / 1000;

  map<UInt32, UInt32>::iterator b = buckets.find(n);
  if (b == buckets.end())
    buckets[n] = 1;
  else
    b->second++;
}


// Add in the contents of another histogram
void Histogram::addHist(const Histogram *hP)
{
  map<UInt32, UInt32>::const_iterator hiter;
  map<UInt32, UInt32>::iterator b;
  for (hiter = hP->buckets.begin(); hiter != hP->buckets.end(); ++hiter)
  {
    b = buckets.find(hiter->first);
    if (b == buckets.end())
      buckets[hiter->first] = hiter->second;
    else
      b->second += hiter->second;
  }
  nEvents += hP->nEvents;
  totalTime += hP->totalTime;
}


// Print a histogram
ostream &operator<<(ostream &os, const Histogram &h)
{
  h.printHist(os);
  return os;
}


// Print a histogram
void Histogram::printHist(ostream &os) const
{
  int oldwidth;
  map<UInt32, UInt32>::const_iterator b;

  if (nEvents == 0)
    return;

  oldwidth = os.width();
  for (b = buckets.begin(); b != buckets.end(); ++b)
    os << "  " << setw(4) << b->first << "  " << setw(7) << b->second << endl;
  os.width(oldwidth);
}


// Return median value in histogram
UInt32 Histogram::median() const
{
  int n = 0, retval = 0;
  map<UInt32, UInt32>::const_iterator b;
  for (b = buckets.begin(); b != buckets.end(); ++b)
  {
    n += b->second;
    if (n >= nEvents/2)
    {
      retval = b->first;
      break;
    }
  }
  return retval;
}


// Return maximum bucket value in this histogram
UInt32 Histogram::maxBucket() const
{
  if (buckets.empty())
    return 0;
  return max_element(buckets.begin(), buckets.end(), BucketCmp())->second;
}


// Send a reply to a message with data from the specified DataBuff.  If the
// DataBuff pointer is NULL, then send an empty reply.  If errText is not
// empty then send an error reply (mtReplyErr) with that text.
void RcvMsg::sendReply(DataBuff *dbP, string errText)
{
  if (!errText.empty())
  {
    msgBuff.newBuff(calcLen(errText));
    msgBuff.putString(errText);
    connP->sendit(mtReplyErr, msgType, msgId, msgBuffP(), msgLen(), NULL);
  }
  else if (dbP == NULL)
    connP->sendit(mtReply, msgType, msgId, NULL, 0, NULL);
  else
    connP->sendit(mtReply, msgType, msgId, dbP->getBuffP(),
                  dbP->getBufflen(), NULL);
}


// If an error occurred on this reply, print a message and return true
bool RcvMsg::showError()
{
  if (errText.empty())
    return false;
  Log("Error reply from " << connP->destName() << ": " << errText);
  return true;
}


// Start handling a request from the admin node.  If another request is
// already being handled, send back an error reply and return true.  The
// admin node should never send a new request before the previous one has
// finished.  This routine is to catch the case of multiple admin nodes, or
// an admin node restart.
bool RcvMsg::startAdminReq()
{
  if (!server)
    Error("admin request received from " << connP->destName());
  thLock(&globalMutex);
  if (cmdInProgress)
  {
    thUnlock(&globalMutex);
    sendReply(NULL, "command already in progress");
    return true;
  }
  cmdInProgress = true;
  thUnlock(&globalMutex);
  return false;
}


// Stop handling an admin command request
void RcvMsg::endAdminReq()
{
  thLock(&globalMutex);
  cmdInProgress = false;
  thUnlock(&globalMutex);
}


// Message handler for mtVersion
//    Input:
//       None
//    Returns:
//       String version
void RcvMsg::handleVersion()
{
  msgBuff.newBuff(calcLen(version));
  msgBuff.putString(version);
  sendReply(&msgBuff);
}


// Message handler for mtNwrite
//    Input:
//       None
//    Returns:
//       Nothing
void RcvMsg::handleNwrite()
{
  Errno err;
  MsgRecord mr;
  string errText;

  // This is an NSD-style write request.  We are supposed to allocate a
  // buffer, tell the other node to send the data to us (targeted to that
  // buffer), and then reply.  For test purposes, we don't care what he
  // sends, so just use a read request.
  err = connP->sendMessage(mtRead, NULL, 0, &mr);
  mr.waitForReplies();
  if (err != E_OK)
  {
    sendReply(NULL, "send failed");
    return;
  }
  RcvMsg *replyP = mr.nextReply();
  errText = replyP->errText;
  delete replyP;

  sendReply(NULL, errText);
}


// Message handler for mtConnect
//    Input:
//       UInt32 nServers
//       Array[nServers]
//         String hostname
//         IpAddr IP address (native host format)
//    Returns:
//       Nothing
void RcvMsg::handleConnect()
{
  int j, nServers = 0;
  IpAddr iaddr;
  string hname, errmsg;
  multimap<IpAddr, Target *>::iterator node;
  Target *targP;
  ostringstream os;

  if (startAdminReq())
    return;

  // Target nodes shouldn't have a client table; that's only for admin
  // nodes.
  if (!clientNodes.empty() || !allNodes.empty())
    Error("nodes table not empty");

  // Remove any broken connections from server table.  Mark the other nodes
  // inactive initially.
  for (node = serverNodes.begin(); node != serverNodes.end(); )
  {
    targP = node->second;
    if (targP->connP == NULL || targP->connP->isBroken())
    {
      delete targP;
      serverNodes.erase(node++);
    }
    else
    {
      targP->active = false;
      ++node;
    }
  }

  nServers = msgBuff.getUInt32();
  for (j = 0; j < nServers; j++)
  {
    hname = msgBuff.getString();
    iaddr = msgBuff.getIpAddr();

    // If this node is already in the server table, then we don't need to
    // connect to it here.  Mark it active so that it won't be deleted.
    // If we have multiple connections to the same IP address, only mark
    // the first "nParallel" connections active.
    pair<multimap<IpAddr, Target *>::iterator,
         multimap<IpAddr, Target *>::iterator> nodes;
    int nConnections = 0;
    nodes = serverNodes.equal_range(iaddr);
    for (node = nodes.first; node != nodes.second; ++node)
    {
      if (nConnections < nParallel)
        node->second->active = true;
      nConnections++;
    }
    if (nConnections > 0)
      continue;

    // This node isn't in table, so connect to it and add to table
    targP = new Target(hname, iaddr);
    errmsg = targP->makeConnection();
    if (!errmsg.empty())
    {
      os << "Cannot connect to " << hname << ": " << errmsg << endl;
      delete targP;
    }
    else
      serverNodes.insert(pair<IpAddr, Target *>(iaddr, targP));
  }

  // Shut down any connections to servers that were not mentioned in the
  // message argument or exceeded the parallel count.
  for (node = serverNodes.begin(); node != serverNodes.end(); )
  {
    targP = node->second;
    if (targP->active)
      ++node;
    else
    {
      delete targP;
      serverNodes.erase(node++);
    }
  }

  // If more than one parallel connection was requested, make additional
  // connections if necessary.  They will look like extra serverNodes.
  if (nParallel > 1)
  {
    multimap<IpAddr, Target *>::const_iterator node;
    map<IpAddr, Target *> servers;
    map<IpAddr, Target *>::const_iterator srv;

    // Make a table of unique server IP addresses
    for (node = serverNodes.begin(); node != serverNodes.end(); ++node)
      servers[node->second->iaddr] = node->second;

    // For each unique address, count the number of existing connections
    // and add more if necessary.
    for (srv = servers.begin(); srv != servers.end(); ++srv)
      for (j = serverNodes.count(srv->first); j < nParallel; j++)
      {
        targP = new Target(srv->second->hostname, srv->first);
        errmsg = targP->makeConnection();
        if (!errmsg.empty())
        {
          os << "Cannot make parallel connection to " << srv->second->hostname
             << ": " << errmsg << endl;
          delete targP;
        }
        else
          serverNodes.insert(pair<IpAddr, Target *>(srv->first, targP));
      }
  }

  endAdminReq();
  sendReply(NULL, os.str());
}


// Message handler for mtReset
//    Input:
//       None
//    Returns:
//       Nothing
void RcvMsg::handleReset()
{
  multimap<IpAddr, Target *>::iterator node;

  if (startAdminReq())
    return;
  if (!clientNodes.empty() || !allNodes.empty())
    Error("nodes table not empty");
  for (node = serverNodes.begin(); node != serverNodes.end(); )
  {
    delete node->second;
    serverNodes.erase(node++);
  }
  endAdminReq();
  sendReply(NULL);
}


// Fill test buffer with random data.  Size is taken from buffsize.
static void fillBuff()
{
  unsigned int oldlen = testBuff.getBufflen();

  // Set assigned buffer length, which could be less than the amount
  // allocated if we are re-using a bigger buffer.
  testBuff.newBuff(buffsize);

  // Don't initialize if we already had a big enough buffer
  if (oldlen < buffsize)
  {
    unsigned int j;
    char *p = testBuff.getBuffP();
    for (j = 0; j < buffsize; j++)
      *p++ = random() % 255;
  }
}


// Break any existing TCP connection for this target
static void disconnect(pair<IpAddr, Target *> p)
{
  Target *targP = p.second;
  if (targP->connP == NULL || targP->connP->isBroken())
    return;
  targP->connP->connShutdown();
  targP->connP->releaseConn();
  targP->connP = NULL;
}


// Message handler for mtParms
//    Input:
//       UInt64 testTime
//       UInt32 buffsize
//       UInt32 socksize
//       UInt32 nTesterThreads
//       UInt32 nParallel
//    Returns:
//       Nothing
void RcvMsg::handleParms()
{
  int newSocksize;
  if (startAdminReq())
    return;
  testTime = msgBuff.getUInt64();
  buffsize = msgBuff.getUInt32();
  newSocksize = msgBuff.getUInt32();
  nTesterThreads = msgBuff.getUInt32();
  nParallel = msgBuff.getUInt32();
  if (testTime < 1) testTime = 1;
  if (buffsize < 1) buffsize = 1;
  if (buffsize > MAX_BUFFSIZE) buffsize = MAX_BUFFSIZE;
  if (newSocksize > MAX_SOCKSIZE) newSocksize = MAX_SOCKSIZE;
  if (nTesterThreads < 1) nTesterThreads = 1;
  if (nTesterThreads > 512) nTesterThreads = 512;
  if (nParallel < 1) nParallel = 1;
  if (nParallel > 128) nParallel = 128;
  fillBuff();

  // If socket buffer size changes, close any existing server connections
  // so that they will be re-opened with the proper settings.  This will
  // also ensure that the server has found out the new socksize value and
  // used that setting on his end.  Also, tell the listen/accept thread to
  // switch to the new size.
  if (newSocksize != socksize)
  {
    socksize = newSocksize;
    for_each(serverNodes.begin(), serverNodes.end(), disconnect);
    if (laThreadP != NULL)
      laThreadP->updateSocksize();
  }
  
  // If we don't have enough tester threads, start up some new ones.
  // If we have too many, wake them up so that some of them will exit.
  thLock(&testerMutex);
  while (testerTab.size() < nTesterThreads)
  {
    thUnlock(&testerMutex);
    Tester *tP = new Tester();
    tP->init();
    thLock(&testerMutex);
    if (!testerTab.insert(tP).second)
      Error("duplicate testerTab entry");
  }
  if (testerTab.size() > nTesterThreads)
  {
    thBcast(&testerCond);
    while (testerTab.size() > nTesterThreads)
      thWait(&testerCond, &testerMutex);
  }
  thUnlock(&testerMutex);

  endAdminReq();
  sendReply(NULL);
}


// Message handler for mtTest
//    Input:
//       UInt32 testType
//    Returns:
//       UInt64 totBytes
//       UInt32 CPU idle percentage
//       Histogram response times
void RcvMsg::handleTest()
{
  TType ttin, tt;
  multimap<IpAddr, Target *>::const_iterator node;
  string errText;
  Histogram hist;

  if (startAdminReq())
    return;
  if (serverNodes.empty())
    errText = "no server nodes for mtTest";

  // Pick up test type from incoming message
  ttin = static_cast<TType>(msgBuff.getUInt32());
  if (ttin >= ttLast && errText.empty())
    errText = "invalid test type";

  if (errText.empty())
  {
    // Run the test.  Pass a TestReq to each tester thread, and wait for
    // them all to complete it.
    HTime endTime;
    UInt64 totBytes;
    TestReq *trP;
    set<Tester *>::iterator tst;

    thLock(&testerMutex);
    if (testerTab.empty())
      Error("no tester threads");
    endTime = getTime() + testTime * 1000000LL;

    // The serverNodes table is ordered by IP address, and might contain
    // more than one connection per address.  We want round-robin sends to
    // go to each server before repeating on parallel connections, so make
    // a new table of addresses with proper sort order.  Note that the
    // serverNodes table will not change in the middle of a test, so we
    // don't need a mutex to access it here.
    vector<SortedTarget> rrNodes;
    IpAddr lastAddr;
    int ndx = 0;

    rrNodes.reserve(serverNodes.size());
    for (node = serverNodes.begin(); node != serverNodes.end(); ++node)
    {
      Target *targP = node->second;
      if (targP->iaddr != lastAddr)
      {
        lastAddr = targP->iaddr;
        ndx = 0;
      }
      rrNodes.push_back(SortedTarget(ndx, targP));
      ndx++;
    }
    sort(rrNodes.begin(), rrNodes.end(), SortedTarget::comp);
    vector<SortedTarget>::const_iterator rnode = rrNodes.begin();
    int nServers = rrNodes.size() / nParallel;

    for (tst = testerTab.begin(); tst != testerTab.end(); ++tst)
    {
      nTestersWorking++;

      // For the read/write test, half the workers are readers and half are
      // writers.
      if (ttin == ttRW)
        tt = nTestersWorking > testerTab.size() / 2 ? ttWrite : ttRead;
      else
        tt = ttin;
      trP = new TestReq(tt, endTime);

      // If test isn't round-robin, assign one server target.
      // Otherwise assign all of them.  With parallel sockets, give
      // separate sets of connections to each thread if possible.
      int ns = (tt == ttSwrite || tt == ttSread) ? 1 : nServers;
      for (ndx = 0; ndx < ns; ndx++)
      {
        trP->testNodes.push_back(rnode->targP);
        rnode++;
        if (rnode == rrNodes.end())
          rnode = rrNodes.begin();
      }

      // Hand off the test request to a tester thread
      (*tst)->doTest(trP);
    }
    thBcast(&testerCond);

    while (nTestersWorking > 0)
      thWait(&testerCond, &testerMutex);

    // Tally the results from the tester threads
    totBytes = 0;
    while (!doneList.empty())
    {
      trP = doneList.front();
      doneList.pop_front();
      totBytes += trP->totBytes;
      hist.addHist(&trP->hist);
      if (errText.empty())
        errText = trP->errText;
      delete trP;
    }
    thUnlock(&testerMutex);

    msgBuff.newBuff(sizeof(UInt64) + sizeof(UInt32) + hist.calcLen());
    msgBuff.putUInt64(totBytes);
    msgBuff.putUInt32(idleTime);
    hist.putBuff(&msgBuff);
  }

  endAdminReq();
  sendReply(&msgBuff, errText);
}


// Message handler for mtStatus
//    Input:
//       None
//    Returns:
//       String ServerStatus
void RcvMsg::handleStatus()
{
  ostringstream os;
  multimap<IpAddr, Target *>::iterator node;

  if (startAdminReq())
    return;
  for (node = serverNodes.begin(); node != serverNodes.end(); ++node)
    os << "    -> " << node->second->name() << endl;

  string s = os.str();
  msgBuff.newBuff(calcLen(s));
  msgBuff.putString(s);
  endAdminReq();
  sendReply(&msgBuff);
}


// Message handler for mtIdlePct
//    Input:
//       None
//    Returns:
//       UInt32 CPU idle percentage
void RcvMsg::handleIdlePct()
{
  msgBuff.newBuff(sizeof(UInt32));
  msgBuff.putUInt32(idleTime);
  sendReply(&msgBuff);
}


// Dispatch a received message to the appropriate handler.  Simple messages
// are handled here.
void RcvMsg::dispatch()
{
  if (debugLevel > 0)
    Logt("got msg type " << mtToString(msgType) << " ID " << msgId
         << " len " << msgLen() << " from " << connP->destName());

  switch (msgType)
  {
    case mtVersion:
      handleVersion();
      break;

    case mtEcho:
      sendReply(&msgBuff);
      break;

    case mtKill:
      sendReply(NULL);
      connP->connShutdown();
      exit(EXIT_SUCCESS);
      break;

    case mtWrite:
      sendReply(NULL);
      break;

    case mtRead:
      sendReply(&testBuff);
      break;

    case mtNwrite:
      handleNwrite();
      break;

    case mtConnect:
      handleConnect();
      break;

    case mtReset:
      handleReset();
      break;

    case mtParms:
      handleParms();
      break;

    case mtTest:
      handleTest();
      break;

    case mtStatus:
      handleStatus();
      break;

    case mtStatOn:
      collectStats = true;
      (void) cpuIdle();
      sendReply(NULL);
      break;

    case mtStatOff:
      collectStats = false;
      idleTime = cpuIdle();
      sendReply(NULL);
      break;

    case mtIdlePct:
      handleIdlePct();
      break;

    default:
      Error("Invalid message from " << connP->destName());
  }
}


// Message worker thread body
int MsgWorker::threadBody()
{
  RcvMsg *rmsgP;
  thLock(&workerMutex);
  while (true)
  {
    // Wait for a receiver thread to pass us a message.  Don't process bulk
    // messages if almost all of the workers are busy.
    while (msgQueue.empty() &&
           (bulkQueue.empty() || workersActive >= nWorkers - 2))
      thWait(&workerCond, &workerMutex);
    workersActive++;
    if (msgQueue.empty())
    {
      rmsgP = bulkQueue.front();
      bulkQueue.pop_front();
    }
    else
    {
      rmsgP = msgQueue.front();
      msgQueue.pop_front();
    }

    // Call the message handler
    thUnlock(&workerMutex);
    rmsgP->dispatch();
    delete rmsgP;
    thLock(&workerMutex);
    workersActive--;
  }
  thUnlock(&workerMutex);
  return 0;
}


// Tester thread body
int Tester::threadBody()
{
  HTime now, lastTime;
  list<Target *>::iterator node;
  Target *targP;
  RcvMsg *replyP = NULL;

  thLock(&testerMutex);
  while (true)
  {
    // Wait for a work request or a notification that too many threads
    // are running.
    while (reqP == NULL && testerTab.size() <= nTesterThreads)
      thWait(&testerCond, &testerMutex);
    if (testerTab.size() > nTesterThreads)
      break;
    thUnlock(&testerMutex);

    if (reqP->testNodes.empty())
      Error("No servers for tester thread");

    // Move to a random starting position in the test nodes list
    int spos;
    node = reqP->testNodes.begin();
    for (spos = random() % reqP->testNodes.size(); spos > 0; spos--)
      if (++node == reqP->testNodes.end())
        node = reqP->testNodes.begin();

    // Run the test
    now = getTime();
    while (now < reqP->endTime)
    {
      lastTime = now;

      targP = *node++;
      if (node == reqP->testNodes.end())
        node = reqP->testNodes.begin();

      switch (reqP->tt)
      {
        case ttWrite:
        case ttSwrite:
          replyP = targP->sendm(mtWrite, testBuff.getBuffP(),
                                testBuff.getBufflen());
          break;
        case ttRead:
        case ttSread:
          replyP = targP->sendm(mtRead, NULL, 0);
          break;
        case ttNwrite:
          replyP = targP->sendm(mtNwrite, NULL, 0);
          break;
        default:
          Error("invalid test type: " << reqP->tt);
          break;
      }
      now = getTime();
      if (!replyP->errText.empty())
      {
        reqP->errText = replyP->errText;
        delete replyP;
        break;
      }
      delete replyP;

      if (collectStats)
      {
        reqP->totBytes += buffsize + MSG_HDRSIZE;
        reqP->hist.addEntry(now - lastTime);
      }
    }
    reqP->endTime = now;

    // Report results to main thread
    thLock(&testerMutex);
    doneList.push_back(reqP);
    reqP = NULL;
    if (nTestersWorking == 0)
      Error("doneList overflow");
    nTestersWorking--;
    if (nTestersWorking == 0)
      thBcast(&testerCond);
  }

  if (testerTab.erase(this) != 1)
    Error("thread pointer not found in testerTab");
  thBcast(&testerCond);
  thUnlock(&testerMutex);
  return 0;
}


// Tell tester thread to do some work.  This doesn't wake up the thread.
// Caller must broadcast testerCond to do that.  Tester will put updated
// TestReq on doneList when finished.  Caller must hold testerMutex.
void Tester::doTest(TestReq *trP)
{
  if (reqP != NULL)
    Error("tester thread busy");
  reqP = trP;
}


// Listen/accept thread body
int ListenAccept::threadBody()
{
  int rc;
  Sock lsock, sock;
  struct linger ling;
#ifdef IPV6_SUPPORT
  struct sockaddr_in6 saddr6;
#endif
  struct sockaddr_in saddr4;
  struct sockaddr *saddrP;
  socklen_t saddrLen, alen;
  PollSock ps;
  IpAddr iaddr;

  // Loop, opening a listen socket
  while (true)
  {
    if ((lsock = socket(addrFamily, SOCK_STREAM, 0)) == INVALID_SOCK)
      Errorm("socket open"); 

    // Set to recycle addresses
    if ((setsockopt(lsock, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on))) < 0)
      Errorm("setsockopt SO_REUSEADDR");

    // Set the socket to not linger after the program exits
    ling.l_onoff = 1;
    ling.l_linger = 0;
    if (setsockopt(lsock, SOL_SOCKET, SO_LINGER, &ling, sizeof(ling)) < 0)
      Errorm("setsockopt SO_LINGER");

    // Set send/receive buffer sizes
    setSockSizes(lsock);

    // Bind the socket
#ifdef IPV6_SUPPORT
    if (useipv6)
    {
      memset(&saddr6, 0, sizeof(saddr6));
      saddr6.sin6_family = addrFamily;
      saddr6.sin6_port = htons(port);
      saddr6.sin6_addr = in6addr_any;
      saddrP = reinterpret_cast<sockaddr *>(&saddr6);
      saddrLen = sizeof(saddr6);
    }
    else
#endif
    {
      memset(&saddr4, 0, sizeof(saddr4));
      saddr4.sin_family = addrFamily;
      saddr4.sin_port = htons(port);
      saddr4.sin_addr.s_addr = htonl(INADDR_ANY);
      saddrP = reinterpret_cast<sockaddr *>(&saddr4);
      saddrLen = sizeof(saddr4);
    }
    if (bind(lsock, saddrP, saddrLen) < 0)
      Errorm("bind");
    if (listen(lsock, 128) < 0) Errorm("listen");

    // Set for non-blocking IO
    setSockNonblocking(lsock);
    
    // Tell updateSocksize that we have adjusted the size
    thLock(&globalMutex);
    currsize = socksize;
    thBcast(&globalCond);
    thUnlock(&globalMutex);

    // Loop, accepting connections
    while (true)
    {
      // If socket buffer size changes, close the listen socket and break
      // out of accept loop to go back and re-open socket with correct size
      // (since size must be set before listen call).
      if (socksize != currsize)
      {
        if (close(lsock) < 0) Errorm("close lsock");
        break;
      }

      // Wait for a connection.  Use a timeout on poll, so that if we miss
      // the signal for a socksize change, we'll wake up soon and notice
      // it.  Using ppoll would be better, but not all systems support that.
      ps.fd = lsock;
      ps.events = POLLIN;
      ps.revents = 0;
      rc = poll(&ps, 1, 1000);
      if (rc < 0)
      {
        if (errno == EINTR)
          continue;
        Errorm("poll for incoming connection");
      }
      if (rc == 0)
        continue;

      // Accept anybody who connects
      alen = saddrLen;
      sock = accept(lsock, saddrP, &alen);
      if (sock == INVALID_SOCK)
      {
        if (errno == EAGAIN || errno == EWOULDBLOCK ||
            errno == ECONNABORTED || errno == EINTR)
          continue;
        Errorm("accept");
      }

      // Get peer address
      alen = saddrLen;
      if (getpeername(sock, saddrP, &alen) < 0)
        Errorm("getpeername");
      iaddr.loadSockaddr(saddrP);
      Log("Connection from " << iaddr.tostring());

      // Don't delay sending messages
      if (setsockopt(sock, IPPROTO_TCP, TCP_NODELAY, &on, sizeof(on)) < 0)
        Errorm("setsockopt TCP_NODELAY");

      // Set for non-blocking IO
      setSockNonblocking(sock);
      
      // Hand off the socket to a receiver thread
      pickReceiver()->addConn(sock, new TcpConn(sock, iaddr));
    }
  }
  return 0;
}


// Tell the listen/accept thread that the socket send/receive buffer size
// has changed, so the listen socket must be closed and re-opened to set
// the new size.
void ListenAccept::updateSocksize()
{
  if (socksize == currsize)
    return;

  // Wake up the thread, if it is in poll.  If it is just about to start
  // the poll, it could miss this signal, but it will time out soon after
  // that and notice the change.
  if (pthread_kill(getThread(), SIGUSR1) < 0) Errorm("pthread_kill");

  // Wait until the thread has picked up the new size.
  thLock(&globalMutex);
  while (socksize != currsize)
    thWait(&globalCond, &globalMutex);
  thUnlock(&globalMutex);
}


// Print help message
static void helpCmd(vector<string> *argsP)
{
  cout << "Commands:"
       << endl
       << "  help                  Print help message" << endl
       << "  quit                  Exit from program" << endl
       << "  version               Show program version" << endl
       << "  debug [LEVEL]         Set debugging output level" << endl
       << "  source FILENAME...    Read commands from file" << endl
       << "  check HOSTNAME...     Quick performance test to single nodes from local" << endl
       << "  server HOSTNAME...    Designate nodes as servers" << endl
       << "  client HOSTNAME...    Designate nodes as clients" << endl
       << "  delete HOSTNAME...    Remove server or client nodes from test" << endl
       << "  reset                 Close connections and clear node tables" << endl
       << "  status                Show status of all nodes" << endl
       << "  ttime NSEC            Set run time (in seconds) for tests" << endl
       << "  buffsize NBYTES       Set I/O data buffer size for tests" << endl
       << "  socksize NBYTES       Set size of TCP send/receive buffer space" << endl
       << "  threads NTHREADS      Set number of tester threads to use on clients" << endl
       << "                          (default " << TESTER_THREADS << ")" << endl
       << "  parallel N            Set number of parallel socket connections (default 1)" << endl
       << "  hist [on|off]         Set printing of response time histograms on or off" << endl
       << "                          (if no option given, then toggle)" << endl
       << "  plot [FILENAME]       Write test results in gnuplot format to specified file" << endl
       << "                          (if no filename given, then turn off plotting)" << endl
       << "  test TTYPE...         Run performance tests from clients to servers" << endl
       << "  kill HOSTNAME...      Shut down " << progname << " server on specified nodes" << endl
       << "  killall               Shut down " << progname << " on all client and server nodes" << endl
       << endl
       << "Test types:" << endl
       << "  write         Clients write round-robin to all servers" << endl
       << "  read          Clients read round-robin from all servers" << endl
       << "  nwrite        Same as write test, but use NSD-style writing" << endl
       << "  swrite        Each tester thread writes to only one server" << endl
       << "  sread         Each tester thread reads from only one server" << endl
       << "  rw            Half of the tester threads read and half write" << endl;
}


// Exit from program
static void quitCmd(vector<string> *argsP)
{
  quitflag = true;
}


// Read commands from file
static void sourceCmd(vector<string> *argsP)
{
  vector<string>::iterator arg;
  ifstream infile;
  string fname, line;

  if (argsP->empty())
  {
    Log("File name missing");
    return;
  }
  nestingLevel++;
  if (nestingLevel > 10)
  {
    Log("Nesting level too deep");
    return;
  }
  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    fname = *arg;
    infile.open(fname.c_str());
    if (!infile)
    {
      Logm("open of " << fname << " failed");
      break;
    }
    while (getline(infile, line))
      splitCmd(line, &commands);
    infile.close();
  }
}


// Show program version
static void versionCmd(vector<string> *argsP)
{
  Log(version);
}


// Set debugging output level
static void debugCmd(vector<string> *argsP)
{
  if (argsP->empty())
    debugLevel++;
  else
    debugLevel = atoi((*argsP)[0].c_str());
  Log("Debugging output level set to " << debugLevel);
}


// Send current test parameters to target
static void sendParms(Target *targP)
{
  DataBuff db(sizeof(UInt64) + 4 * sizeof(UInt32));
  db.putUInt64(testTime);
  db.putUInt32(buffsize);
  db.putUInt32(socksize);
  db.putUInt32(nTesterThreads);
  db.putUInt32(nParallel);

  RcvMsg *rmsgP = targP->sendm(mtParms, db.getBuffP(), db.getBufflen());
  rmsgP->showError();
  delete rmsgP;
}


// Same as sendParms, but takes a pair as input.  The first member of the
// pair is ignored.
static void sendParms2(pair<IpAddr, Target *> p)
{
  sendParms(p.second);
}

  
// Quick single-threaded performance test to individual nodes from local
static void checkCmd(vector<string> *argsP)
{
  IpAddr iaddr;
  Target *targP;
  string errmsg;
  RcvMsg *rmsgP;
  HTime startTime, endTime, lastTime;
  UInt64 totBytes;
  vector<string>::iterator arg;
  Histogram hist;
  bool gotError;
  list<TType> tests;
  list<TType>::iterator tt;

  if (argsP->empty())
  {
    Log("Host name missing");
    return;
  }

  tests.push_back(ttWrite);
  tests.push_back(ttRead);

  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    if (iaddr.parse(*arg) != E_OK)
      continue;
    targP = new Target(*arg, iaddr);
    errmsg = targP->makeConnection();
    if (!errmsg.empty())
    {
      Log(errmsg);
      delete targP;
      continue;
    }

    // Tell remote node the test parameters
    sendParms(targP);

    for (tt = tests.begin(); tt != tests.end(); ++tt)
    {
      totBytes = 0;
      startTime = endTime = getTime();
      gotError = false;
      while (endTime - startTime < testTime * 1000000LL)
      {
        lastTime = endTime;
        rmsgP = targP->sendm(*tt == ttWrite ? mtWrite : mtRead,
                             testBuff.getBuffP(), testBuff.getBufflen());
        if (rmsgP->showError())
          gotError = true;
        delete rmsgP;
        if (gotError)
          break;

        endTime = getTime();
        hist.addEntry(endTime - lastTime);
        totBytes += buffsize + MSG_HDRSIZE;
      }
      if (gotError)
      {
        delete targP;
        continue;
      }

      double elapsedT = (endTime - startTime) / 1000000.0;
      double rate = siground(totBytes / elapsedT / 1000000.0, 3);
      if (*tt == ttWrite)
        Log("write rate " << rate << " MB/sec");
      else
        Log("read rate " << rate << " MB/sec");
      if (showHist)
        cout << "block transmit times in msec (average "
             << siground(hist.average() / 1000.0, 4)
             << ", median " << hist.median() << ")"
             << "  msec  nevents" << endl
             << hist;
    }
    delete targP;
  }
}


// Close connections and clear node tables
static void resetCmd(vector<string> *argsP)
{
  map<IpAddr, Target *>::iterator node;

  // Tell every node that we know about to close any connections that
  // it has, and then remove the node from local tables.
  for (node = allNodes.begin(); node != allNodes.end(); )
  {
    Target *targP = node->second;
    RcvMsg *rmsgP = targP->sendm(mtReset, NULL, 0);
    rmsgP->showError();
    delete rmsgP;
    delete targP;
    allNodes.erase(node++);
  }
  serverNodes.clear();
  clientNodes.clear();
}


// Shut down all client and server nodes
static void killallCmd(vector<string> *argsP)
{
  map<IpAddr, Target *>::iterator node;
  for (node = allNodes.begin(); node != allNodes.end(); )
  {
    Target *targP = node->second;
    delete targP->sendm(mtKill, NULL, 0);
    delete targP;
    allNodes.erase(node++);
  }
  serverNodes.clear();
  clientNodes.clear();
}


// Shut down server on specified nodes
static void killCmd(vector<string> *argsP)
{
  IpAddr iaddr;
  Target *targP;
  string errmsg;
  vector<string>::iterator arg;

  if (argsP->empty())
  {
    Log("Host name missing");
    return;
  }
  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    if (iaddr.parse(*arg) != E_OK)
      continue;
    targP = new Target(*arg, iaddr);
    errmsg = targP->makeConnection();
    if (!errmsg.empty())
      Log("Cannot connect to " << *arg << ": " << errmsg);
    else
      delete targP->sendm(mtKill, NULL, 0);
    delete targP;
  }
}


// Common routine for serverCmd and clientCmd
static void doClientServer(vector<string> *argsP, list<Target *> *targListP)
{
  string hname, errmsg;
  IpAddr iaddr;
  Target *targP;
  vector<string>::iterator arg;

  if (argsP->empty())
  {
    Log("Host name missing");
    return;
  }
  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    hname = *arg;
    if (iaddr.parse(hname) != E_OK)
      continue;
    if (allNodes.find(iaddr) != allNodes.end())
    {
      Log(hostString(hname, iaddr) << " is already in use");
      continue;
    }
    targP = new Target(hname, iaddr);
    errmsg = targP->makeConnection();
    if (!errmsg.empty())
    {
      Log("Cannot connect to " << hname << ": " << errmsg);
      delete targP;
    }
    else
      targListP->push_back(targP);
  }
}


// Designate nodes as servers
static void serverCmd(vector<string> *argsP)
{
  list<Target *> targList;
  doClientServer(argsP, &targList);

  list<Target *>::const_iterator targ;
  for (targ = targList.begin(); targ != targList.end(); ++targ)
  {
    serverNodes.insert(pair<IpAddr, Target *>((*targ)->iaddr, *targ));
    allNodes[(*targ)->iaddr] = *targ;
  }
}


// Designate nodes as clients
static void clientCmd(vector<string> *argsP)
{
  list<Target *> targList;
  doClientServer(argsP, &targList);

  list<Target *>::const_iterator targ;
  for (targ = targList.begin(); targ != targList.end(); ++targ)
  {
    clientNodes[(*targ)->iaddr] = *targ;
    allNodes[(*targ)->iaddr] = *targ;
  }
}


// Remove server or client nodes from test
static void deleteCmd(vector<string> *argsP)
{
  string hname;
  IpAddr iaddr;
  Target *targP;
  vector<string>::iterator arg;
  map<IpAddr, Target *>::iterator node;

  if (argsP->empty())
  {
    Log("Host name missing");
    return;
  }
  for (arg = argsP->begin(); arg != argsP->end(); ++arg)
  {
    hname = *arg;
    if (iaddr.parse(hname) != E_OK)
      return;
    node = allNodes.find(iaddr);
    if (node == allNodes.end())
    {
      Log(hostString(hname, iaddr) << " not found");
      continue;
    }
    targP = node->second;

    // Don't worry if send fails.  If we ever use this node again, we'll
    // send another mtReset and check results from that.
    if (targP->connP != NULL)
      delete targP->sendm(mtReset, NULL, 0);

    delete targP;
    allNodes.erase(node);
    serverNodes.erase(iaddr);
    clientNodes.erase(iaddr);
  }
}


// Show status of a target
static void showStatus(pair<IpAddr, Target *> p)
{
  Target *targP = p.second;
  cout << "  " << targP->name();
  if (targP->connP == NULL)
  {
    cout << ", not connected" << endl;
    return;
  }
  cout << endl;

  RcvMsg *rmsgP = targP->sendm(mtStatus, NULL, 0);
  if (!rmsgP->errText.empty())
  {
    Log("    " << rmsgP->errText);
    delete rmsgP;
    return;
  }
  cout << rmsgP->msgBuff.getString();
  delete rmsgP;
}


// If the TCP connection for this target is not connected, try to reconnect it
static void reconnect(pair<IpAddr, Target *> p)
{
  Target *targP = p.second;
  if (targP->connP != NULL)
  {
    if (!targP->connP->isBroken())
      return;
    targP->connP->connShutdown();
    targP->connP->releaseConn();
    targP->connP = NULL;
  }
  (void) targP->makeConnection();
}


// Show status of all nodes
static void statusCmd(vector<string> *argsP)
{
  cout << "test time: " << testTime << " sec" << endl
       << "data buffer size: " << buffsize << endl
       << "TCP socket send/receive buffer size: " << socksize << endl
       << "tester threads: " << nTesterThreads << endl
       << "parallel connections: " << nParallel << endl
       << endl;

  if (!clientNodes.empty())
  {
    Log("clients:");
    for_each(clientNodes.begin(), clientNodes.end(), reconnect);
    for_each(clientNodes.begin(), clientNodes.end(), showStatus);
  }
  if (!serverNodes.empty())
  {
    Log("servers:");
    for_each(serverNodes.begin(), serverNodes.end(), reconnect);
    for_each(serverNodes.begin(), serverNodes.end(), showStatus);
  }
}


// Set run time for tests
static void ttimeCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Number of seconds argument missing");
    return;
  }
  testTime = atoi((*argsP)[0].c_str());
  if (testTime < 1)
    testTime = 1;
  Log("Test time set to " << testTime << " seconds");
}


// Set buffer size for tests
static void buffsizeCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Buffer size argument missing");
    return;
  }
  buffsize = atoi((*argsP)[0].c_str());
  if (buffsize < 1) buffsize = 1;
  if (buffsize > MAX_BUFFSIZE) buffsize = MAX_BUFFSIZE;
  fillBuff();
  Log("Buffer size set to " << buffsize << " bytes");
}


// Set size of TCP send/receive buffer space
static void socksizeCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Size argument missing");
    return;
  }
  socksize = atoi((*argsP)[0].c_str());
  if (socksize < 0) socksize = 0;
  if (socksize > MAX_SOCKSIZE) socksize = MAX_SOCKSIZE;
  Log("TCP send/receive buffer size set to " << socksize << " bytes");
}


// Set number of tester threads to use on clients
static void threadsCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Number of threads argument missing");
    return;
  }
  nTesterThreads = atoi((*argsP)[0].c_str());
  if (nTesterThreads < 1)   nTesterThreads = 1;
  if (nTesterThreads > 512) nTesterThreads = 512;
  Log("Number of tester threads set to " << nTesterThreads);
}


// Tell clients to connect to servers
static Errno doConnect()
{
  Errno err = E_OK;
  map<IpAddr, Target *>::const_iterator node;
  multimap<IpAddr, Target *>::const_iterator snode;
  MsgRecord mr;
  RcvMsg *rmsgP;
  DataBuff db;

  if (serverNodes.empty())
  {
    Log("No server nodes found");
    return E_INVAL;
  }
  if (clientNodes.empty())
  {
    Log("No client nodes found");
    return E_INVAL;
  }
  
  // Tell the server nodes to close any connections that they might have
  // had from previous sessions, where they were clients.
  for (snode = serverNodes.begin(); snode != serverNodes.end(); ++snode)
    if (snode->second->connP->sendMessage(mtReset, NULL, 0, &mr) != E_OK)
    {
      Log("Send to " << snode->second->hostname << " failed");
      err = E_SENDFAILED;
    }
  if (mr.checkReplies())
    return E_REPLY;

  // Make a list of all of the servers to send to the clients
  int len = sizeof(UInt32);
  for (snode = serverNodes.begin(); snode != serverNodes.end(); ++snode)
    len += calcLen(snode->second->hostname) + sizeof(UInt32)
      + sizeof(snode->second->iaddr.a);
  db.newBuff(len);
  db.putUInt32(serverNodes.size());
  for (snode = serverNodes.begin(); snode != serverNodes.end(); ++snode)
  {
    db.putString(snode->second->hostname);
    db.putIpAddr(snode->second->iaddr);
  }

  // Send the server list to all clients
  for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
    if (node->second->connP->sendMessage(mtConnect, db.getBuffP(),
                                         db.getBufflen(), &mr) != E_OK)
      Log("Send to " << node->second->hostname << " failed");
  mr.waitForReplies();

  while (true)
  {
    rmsgP = mr.nextReply();
    if (rmsgP == NULL)
      break;
    if (!rmsgP->showError() && debugLevel > 0)
      Log(rmsgP->connP->destName() << " connected");
    delete rmsgP;
  }
  return E_OK;
}


// Run performance test from clients to servers
static void testCmd(vector<string> *argsP)
{
  vector<string>::iterator arg;
  map<IpAddr, Target *>::iterator node;
  list<TType> tests;
  list<TType>::iterator tt;
  MsgRecord mr, mrstat;
  HTime startTime, endTime, runTime, totTime;
  RcvMsg *rmsgP = NULL;
  DataBuff db;
  bool gotIdle;
  UInt32 idlePct;
  UInt64 totBytes, clientIdle, serverIdle;
  Histogram *hP;
  ofstream plfile;
  bool plotOpened = false;

  if (argsP->empty())
  {
    tests.push_back(ttWrite);
    tests.push_back(ttRead);
  }
  else
  {
    for (arg = argsP->begin(); arg != argsP->end(); ++arg)
    {
      if (match("write", *arg))
        tests.push_back(ttWrite);
      else if (match("read", *arg))
        tests.push_back(ttRead);
      else if (match("nwrite", *arg))
        tests.push_back(ttNwrite);
      else if (match("rw", *arg))
      {
        if (nTesterThreads < 2)
        {
          Log("Need at least two tester threads for rw test");
          return;
        }
        tests.push_back(ttRW);
      }
      else if (match("swrite", *arg))
        tests.push_back(ttSwrite);
      else if (match("sread", *arg))
        tests.push_back(ttSread);
      else
      {
        Log("Invalid test type: " << *arg);
        Log("Valid types are: write nwrite read rw swrite sread");
        return;
      }
    }
  }

  if (clientNodes.empty())
  {
    Log("No client nodes found");
    return;
  }
  for_each(allNodes.begin(), allNodes.end(), reconnect);
  for (node = allNodes.begin(); node != allNodes.end(); ++node)
    if (node->second->connP == NULL)
    {
      Log(node->second->name() << " is not connected");
      return;
    }

  // Tell everyone the current test parameters
  for_each(allNodes.begin(), allNodes.end(), sendParms2);

  // Pass list of servers to all clients and have them connect
  if (doConnect() != E_OK)
    return;

  if (!plotFname.empty())
  {
    plfile.open(plotFname.c_str());
    if (!plfile)
    {
      Logm("open of " << plotFname << " failed");
      goto exit;
    }
    plotOpened = true;
  }

  db.newBuff(sizeof(UInt32));
  for (tt = tests.begin(); tt != tests.end(); ++tt)
  {
    startTime = getTime();
    runTime = testTime * 1000000LL;

    // Tell all clients to start the test, but don't start collecting
    // statistics yet.
    db.resetBuff();
    db.putUInt32(*tt);

    for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
      if (node->second->connP->sendMessage(mtTest, db.getBuffP(),
                                           db.getBufflen(), &mr) != E_OK)
      {
        Log("Send to " << node->second->name() << " failed");
        goto exit;
      }

    // Delay for 10% of test time and then tell all nodes to start
    // gathering statistics.
    totTime = sleepUntil(startTime + runTime/10);
    for (node = allNodes.begin(); node != allNodes.end(); ++node)
    {
      Target *targP = node->second;
      if (targP->connP->sendMessage(mtStatOn, NULL, 0, &mrstat) != E_OK)
      {
        Log("Send to " << targP->name() << " failed");
        goto exit;
      }
    }
    if (mrstat.checkReplies())
      goto exit;

    // Stop gathering statistics after 90% of test time.  Verify that we
    // didn't use all of the test time getting started.  We need at least
    // 100 msec left over to run the test
    endTime = startTime + runTime - runTime/10;
    if (endTime - getTime() < 100000)
    {
      Log("Test startup took too long");
      goto exit;
    }
    sleepUntil(endTime);
    for (node = allNodes.begin(); node != allNodes.end(); ++node)
    {
      Target *targP = node->second;
      if (targP->connP->sendMessage(mtStatOff, NULL, 0, &mrstat) != E_OK)
      {
        Log("Send to " << targP->name() << " failed");
        goto exit;
      }
    }
    if (mrstat.checkReplies())
      goto exit;
    totTime = getTime() - totTime;

    // Wait for all clients to finish testing
    mr.waitForReplies();

    // Add up test results
    totBytes = clientIdle = 0;
    gotIdle = true;
    while (true)
    {
      rmsgP = mr.nextReply();
      if (rmsgP == NULL)
        break;
      if (rmsgP->showError())
        goto exit;

      totBytes += rmsgP->msgBuff.getUInt64();
      idlePct = rmsgP->msgBuff.getUInt32();
      if (idlePct <= 100)
        clientIdle += idlePct;
      else
        gotIdle = false;
      rmsgP->connP->getHistP()->getBuff(&rmsgP->msgBuff);
      delete rmsgP;
    }

    // Gather server CPU idle percentage
    multimap<IpAddr, Target *>::const_iterator snode;
    for (snode = serverNodes.begin(); snode != serverNodes.end(); ++snode)
      if (snode->second->connP->sendMessage(mtIdlePct, NULL, 0, &mr) != E_OK)
      {
        Log("Send to " << snode->second->name() << " failed");
        goto exit;
      }
    mr.waitForReplies();

    serverIdle = 0;
    while (true)
    {
      rmsgP = mr.nextReply();
      if (rmsgP == NULL)
        break;
      if (rmsgP->showError())
        goto exit;

      idlePct = rmsgP->msgBuff.getUInt32();
      if (idlePct <= 100)
        serverIdle += idlePct;
      else
        gotIdle = false;
      delete rmsgP;
    }

    // Print results
    cout << clientNodes.size() << "-" << serverNodes.size() << " ";
    switch (*tt)
    {
      case ttWrite:  cout << "write";  break;
      case ttNwrite: cout << "nwrite"; break;
      case ttRead:   cout << "read";   break;
      case ttRW:     cout << "rw";     break;
      case ttSwrite: cout << "swrite"; break;
      case ttSread:  cout << "sread";  break;
      default:       Error("invalid test type"); break;
    }
    double elapsedT = totTime / 1000000.0;
    double rate = siground(totBytes / elapsedT / 1000000.0, 3);
    cout << " " << rate << " MB/sec";
    if (gotIdle)
    {
      cout << ", cli " << 100 - clientIdle / clientNodes.size() << "%";
      cout << " srv " << 100 - serverIdle / serverNodes.size() << "%";
    }
    cout << ", time " << testTime << ", buff " << buffsize;
    if (socksize != 0)
      cout << ", sock " << socksize;
    if (nTesterThreads != TESTER_THREADS)
      cout << ", th " << nTesterThreads;
    if (nParallel > 1)
      cout << ", parallel " << nParallel;
    cout << endl;

    if (showHist)
    {
      cout << endl;
      for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
      {
        Histogram *hP = node->second->connP->getHistP();
        cout << node->second->hostname << " block transmit times in msec ("
             << "average " << siground(hP->average() / 1000.0, 4)
             << ", median " << hP->median() << ")" << endl
             << "  msec  nevents" << endl
             << *hP << endl;
      }
    }

    if (!plotFname.empty())
    {
      unsigned int n = clientNodes.size();
      if (n > 1)
      {
        int rows, cols;
        rows = (n + 3) / 4;
        cols = n > 4 ? 4 : n;
        plfile << "set multiplot layout " << rows << "," << cols
               << " title \"Block transmit times (msec)\"" << endl
               << "set xlabel" << endl
               << "set ylabel" << endl;
      }
      else
        plfile << "set xlabel \"Time (msec)\"" << endl
               << "set ylabel" << endl;

      plfile << "set format y \"\"" << endl
             << "set style data boxes" << endl
             << "set style fill solid" << endl
             << "set boxwidth 0.9" << endl
             << "unset key" << endl
             << "set border 1" << endl
             << "set xtics out nomirror offset 0,0.5" << endl
             << "unset ytics" << endl;

      // Compute bounds for all nodes and use that to set xrange and yrange
      // so that all plots will have the same scale.
      UInt32 b, maxval = 0, minT = 0xFFFFFFFFU, maxT = 0;
      for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
      {
        hP = node->second->connP->getHistP();
        b = hP->maxBucket(); if (b > maxval) maxval = b;
        b = hP->minTime();   if (b < minT)   minT = b;
        b = hP->maxTime();   if (b > maxT)   maxT = b;
      }
      maxT = (maxT + 10) / 10 * 10;
      minT = (minT < 1) ? minT : (minT - 1) / 10 * 10;
      maxval = (maxval + 9) / 10 * 10;

      plfile << "set xrange [" << minT << ":" << maxT << "]" << endl
             << "set yrange [0:" << maxval << "]" << endl;

      for (node = clientNodes.begin(); node != clientNodes.end(); ++node)
      {
        hP = node->second->connP->getHistP();
        plfile << "set label \"" << node->second->hostname << "\""
               << " at graph 0.5,0.8 center" << endl
               << "plot '-' using 1:2" << endl
               << *hP << "e" << endl
               << "unset label" << endl;
      }
      if (clientNodes.size() > 1)
        plfile << "unset multiplot" << endl;

      plfile << "pause mouse any keypress" << endl;
    }
  }

exit:
  // If we broke out early due to an error, clean out any outstanding
  // replies.
  delete rmsgP;
  mrstat.checkReplies();
  mr.checkReplies();
  if (plotOpened)
  {
    plfile.close();
    if (!plfile)
      Logm("close of " << plotFname << " failed");
  }
}


// Toggle printing of client resonse time histograms
static void histCmd(vector<string> *argsP)
{
  if (argsP->empty())
    showHist = !showHist;
  else if ((*argsP)[0] == "on")
    showHist = true;
  else if ((*argsP)[0] == "off")
    showHist = false;
  else
  {
    Log("Invalid option");
    return;
  }
  cout << "Histogram printing is now ";
  if (showHist) cout << "on"; else cout << "off";
  cout << endl;
}


// Set number of parallel socket connections
static void parallelCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    Log("Number of connections argument is missing");
    return;
  }
  nParallel = atoi((*argsP)[0].c_str());
  if (nParallel < 1) nParallel = 1;
  if (nParallel > 128) nParallel = 128;
  Log("Number of parallel socket connections set to " << nParallel);
}


// Write test results in gnuplot format to specified file
static void plotCmd(vector<string> *argsP)
{
  if (argsP->empty())
  {
    plotFname.clear();
    Log("Plotting disabled");
    return;
  }
  plotFname = (*argsP)[0];
  Log("Plot commands will be written to " << plotFname);
}


// Entry in command table
struct Command
{
  string cmd;                                   // Command keyword
  string::size_type minlen;                     // Minium abbreviation
  void (*cmdRtnP)(vector<string> *argsP);       // Routine to call
};

  
// Valid commands
static struct Command cmdTable[] = 
{
  { "hist",     2, histCmd },
  { "help",     1, helpCmd },
  { "?",        1, helpCmd },
  { "quit",     1, quitCmd },
  { "source",   2, sourceCmd },
  { ".",        1, sourceCmd },
  { "version",  1, versionCmd },
  { "debug",    3, debugCmd },
  { "check",    2, checkCmd },
  { "reset",    1, resetCmd },
  { "killall",  5, killallCmd },
  { "ka",       2, killallCmd },
  { "kill",     1, killCmd },
  { "status",   2, statusCmd },
  { "socksize", 2, socksizeCmd },
  { "server",   1, serverCmd },
  { "client",   1, clientCmd },
  { "delete",   1, deleteCmd },
  { "threads",  2, threadsCmd },
  { "ttime",    2, ttimeCmd },
  { "plot",     1, plotCmd },
  { "parallel", 2, parallelCmd },
  { "test",     1, testCmd },
  { "buffsize", 1, buffsizeCmd },
};
static const int nCommands = sizeof(cmdTable) / sizeof(struct Command);  


// Handle SIGUSR1
static void sigusr1Handler(int signum)
{ }


// Handle out of memory error
static void exhausted()
{
  cerr << "Out of storage!" << endl;
  exit(EXIT_FAILURE);
}


// Print usage message and exit
static void usage()
{
  cerr << "usage:  " << progname
       << " [-d] [-h] [-i FNAME] [-p PORT] [-r NRCV] [-s]" << endl
       << "                [-w NWORKERS] [-6] [CMD...]" << endl
       << endl
       << "Options:" << endl
       << "  -d            Include debug output" << endl
       << "  -h            Print help message" << endl
       << "  -i FNAME      Read commands from file FNAME" << endl
       << "  -p PORT       TCP port to use (default " << NSDPERF_PORT << ")" << endl
       << "  -r NRCV       Number of receiver threads (default nCPUs, min 2)" << endl
       << "  -s            Act as a server" << endl
       << "  -w NWORKERS   Number of message worker threads (default " << MSG_WORKERS << ")" << endl
       << "  -6            Use IPv6 rather than IPv4" << endl;

  exit(EXIT_FAILURE);
}


// Start of main program
int main(int argc, char *argvP[])
{
  int j, c;
  string::size_type spos;
  string line;

  // Trap out-of-memory errors
  set_new_handler(exhausted);

  // Get the name of this program for error messages
  progname = argvP[0];
  spos = progname.find_last_of("/\\");
  if (spos != string::npos) progname = progname.substr(spos+1);

  // Figure out the endian-ness of this machine
  union { UInt16 n; UChar c[2]; } endTest;
  endTest.n = 1;
  littleEndian = (endTest.c[0] == 1);

  // Parse options
  string fname;
  int nReceivers = numProcessors();
  while ((c = getopt(argc, argvP, "dhi:p:r:sw:6")) != EOF)
    switch (c)
    {
      case 'd':                 // Include debug output
        debugLevel++;
        break;

      case 'h':                 // Ask for help
        usage();
        break;

      case 'i':                 // Input file name
        fname = optarg;
        break;

      case 'p':                 // Port to use
        port = atoi(optarg);
        break;

      case 'r':                 // Number of receiver threads
        nReceivers = atoi(optarg);
        break;

      case 's':                 // Run in server mode
        server = true;
        break;

      case 'w':                 // Number of message worker threads
        nWorkers = atoi(optarg);
        break;

      case '6':                 // Use IPv6 rather than IPv4
        useipv6 = true;
        break;

      default:  
        exit(EXIT_FAILURE);
        break;
    }

#ifdef IPV6_SUPPORT
  addrFamily = useipv6 ? AF_INET6 : AF_INET;
#else
  if (useipv6) Error("IPv6 not supported on this system");
  addrFamily = AF_INET;
#endif
  if (nReceivers < 2) nReceivers = 2;
  if (nReceivers > 32) nReceivers = 32;
  srandom(12345);
  fillBuff();

  // If input file name was given, read file contents into line buffer
  if (!fname.empty())
  {
    ifstream infile;
    infile.open(fname.c_str());
    if (!infile)
      Errorm("open of " << fname << " failed");
    while (getline(infile, line))
      splitCmd(line, &commands);
    infile.close();
  }

  // Remaining arguments are treated as a command line
  for (j = optind; j < argc; j++)
  {
    if (!line.empty())
      line += " ";
    line += argvP[j];
  }
  splitCmd(line, &commands);
  
  // Set up a signal handler for SIGUSR1
  struct sigaction sa;
  sa.sa_handler = sigusr1Handler;
  sa.sa_flags = 0;
  if (sigaction(SIGUSR1, &sa, NULL) < 0) Errorm("sigaction");

  // Start up threads
  thInitMutex(&globalMutex);
  thInitCond(&globalCond);
  for (j = 0; j < nReceivers; j++)
  {
    Receiver *rcvP = new Receiver;
    rcvP->init();
    receiverTab.push_back(rcvP);
  }
  nextReceiver = receiverTab.begin(); 
  thInitMutex(&workerMutex);
  thInitCond(&workerCond);
  for (j = 0; j < nWorkers; j++)
  {
    MsgWorker *mwP = new MsgWorker();
    mwP->init();
  }

  thInitMutex(&testerMutex);
  thInitCond(&testerCond);

  if (server)
  {
    laThreadP = new ListenAccept;
    laThreadP->init();
    if (debugLevel > 0)
      Log(progname << " server started");

    // Clean up threads if they exit, and quit when all threads are gone
    thLock(&globalMutex);
    while (nThreadsStarted > 0)
    {
      while (deadThreads.empty())
        thWait(&globalCond, &globalMutex);

      Thread *thP = deadThreads.front();
      deadThreads.pop_front();
      thUnlock(&globalMutex);

      thJoin(thP->getThread(), NULL);
      delete thP;

      thLock(&globalMutex);
      nThreadsStarted--;
    }
    thUnlock(&globalMutex);
    exit(EXIT_SUCCESS);
  }

  // Read commands
  bool interactive = isatty(fileno(stdin));
  while (true)
  {
    // If no more pre-read lines, prompt for a new one
    while (commands.empty())
    {
      nestingLevel = 0;
      if (interactive)
        cout << progname << "> ";
      if (!getline(cin, line))
      {
        if (interactive)
          cout << endl;
        quitflag = true;
        break;
      }
      splitCmd(line, &commands);
    }
    if (quitflag)
      break;

    line = commands.front();
    commands.pop_front();

    // First word in line is command name
    spos = 0;
    string cmd = getword(line, spos);
    if (cmd.empty() || match(cmd, "#"))
      continue;

    // Split remainder of line into words, which are the command arguments
    vector<string> args;
    while (spos != string::npos)
    {
      string w = getword(line, spos);
      if (!w.empty())
        args.push_back(w);
    }

    // Look up command in table and run the command routine
    for (j = 0; j < nCommands; j++)
      if (cmd.length() >= cmdTable[j].minlen && match(cmdTable[j].cmd, cmd))
      {
        cmdTable[j].cmdRtnP(&args);
        break;
      }
    if (j == nCommands)
    {
      Log("Invalid command.");
      helpCmd(NULL);
    }
    if (quitflag)
      break;
  }
  exit(EXIT_SUCCESS);
}
