/***************************************************************************
 *
 * Copyright (C) 2001 International Business Machines
 * All rights reserved.
 *
 * This file is part of the GPFS mmfslinux kernel module.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 *  1. Redistributions of source code must retain the above copyright notice, 
 *     this list of conditions and the following disclaimer. 
 *  2. Redistributions in binary form must reproduce the above copyright 
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution. 
 *  3. The name of the author may not be used to endorse or promote products 
 *     derived from this software without specific prior written
 *     permission. 
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR 
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES 
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, 
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, 
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; 
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, 
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF 
 * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *************************************************************************** */
/* @(#)64       1.55  src/avs/fs/mmfs/ts/kernext/gpl-linux/cxiCache.c, mmfs, avs_rche, rche0933a 7/2/09 17:46:51 */
/*
 * Cache (NFS) operations
 *
 * Contents:
 *   cxiCacheInit
 *   cxiCacheExit
 *   cxiCacheLookupRoot
 *   cxiCacheLookup
 *   cxiCacheOpen
 *   cxiCacheClose
 *   cxiCacheCreate
 *   cxiCacheRemove
 *   cxiCacheLink
 *   cxiCacheSymlink
 *   cxiCacheGetattr
 *   cxiCacheSetattr
 *   cxiCacheReaddir
 *   cxiCacheRead
 *   cxiCacheWrite
 *   cxiCacheRename
 *   cxiCacheReadlink
 *   cxiCacheNfsOpen
 *   cxiCacheGetFilename
 */

#include <Shark-gpl.h>
#include <Trace.h>
#include <linux2gpfs.h>

#include <linux/nfs_fs.h>
#include <linux/file.h>
/* #include <linux/fsnotify.h> */

#define PCACHE_XATTR_CTL_FILE   ".pcache.xattr.ctl"

static struct inode *
(*fhget)(struct super_block *, struct nfs_fh *, struct nfs_fattr *) = NULL;

/* Copy NFS filehandle */
static inline void
cache_copy_fh(struct nfs_fh *targetP, const struct nfs_fh *sourceP)
{
  targetP->size = sourceP->size;
  memcpy(targetP->data, sourceP->data, sourceP->size);
}

static inline int
cache_compare_fh(const struct nfs_fh *a, const struct nfs_fh *b)
{
  return a->size != b->size || cxiMemcmp(a->data, b->data, a->size) != 0;
}

/* Convert GPFS vattr to NFS fattr */
static void
vattr2fattr(struct super_block *sbP, cxiVattr_t *vattrP, struct nfs_fattr *fattrP)
{
  u32 time[2];

#if LINUX_KERNEL_VERSION < 2063000
  fattrP->type = vattrP->va_type;
  DBGASSERT(fattrP->type != NFNON);
#endif
  fattrP->valid = NFS_ATTR_FATTR | NFS_ATTR_FATTR_V3;
  if (NFS_PROTO(sbP->s_root->d_inode)->version == 4)
    fattrP->valid |= NFS_ATTR_FATTR_V4;
  fattrP->nlink = vattrP->va_nlink;
  /* FIX: There are cases where we are called for a file that's locally deleted
     (for example to replay a remove). In these cases nlink will be 0 but
     NFS will complain, so we set it to 1 */
  if (fattrP->nlink == 0)
  {
    fattrP->nlink = 1;
  }
  fattrP->fileid = (u64)vattrP->va_ino;
  fattrP->mode = vattrP->va_mode;
  fattrP->size = vattrP->va_size;
#ifdef HAS_NFS_SUBMOUNTS
  /* We don't store server-provided fsid in GPFS, so use the one from the NFS sb */
  memcpy(&NFS_FATTR_FSID(fattrP), &NFS_SB(sbP)->fsid, sizeof(NFS_SB(sbP)->fsid));
  fattrP->time_start = 0;
#else
  NFS_FATTR_FSID(fattrP).major = NFS_FATTR_FSID(fattrP).minor = 0; /* unused */
#endif
  fattrP->rdev = ((vattrP->va_rdev>>16)<<20)|(vattrP->va_rdev & 0xFFFF);
  CXITIME_TO_INODETIME(vattrP->va_atime, fattrP->atime);
  CXITIME_TO_INODETIME(vattrP->va_mtime, fattrP->mtime);
  CXITIME_TO_INODETIME(vattrP->va_ctime, fattrP->ctime);
  time[0] = (u32)vattrP->va_ctime.tv_sec;
  time[1] = (u32)vattrP->va_ctime.tv_nsec;
  memcpy(&fattrP->change_attr, &time, sizeof(u64));
  fattrP->uid = vattrP->va_uid;
  fattrP->gid = vattrP->va_gid;
  fattrP->du.nfs3.used = (u64)vattrP->va_blocks << 9;
}

/* Fill up NFS file attributes from the inode */
static void inode2pattr(struct inode *inodeP, cxiPattr_t *pattrP)
{
  struct nfs_inode *nfsiP = NFS_I(inodeP);
  cxiVattr_t *vattrP = &pattrP->pa_vattr;
  cxiFH_t *fhP = &pattrP->pa_fh;
  __u32 *fh = (__u32 *)fhP->data;

  vattrP->va_mode = inodeP->i_mode;
  vattrP->va_uid = inodeP->i_uid;
  vattrP->va_gid = inodeP->i_gid;
  vattrP->va_rdev = cxiDevToDev32(inodeP->i_rdev);
  vattrP->va_ino = (cxiInoSys_t)nfsiP->fileid;
  vattrP->va_nlink = inodeP->i_nlink;
  vattrP->va_size = inodeP->i_size;
  CXITIME_FROM_INODETIME(vattrP->va_mtime, inodeP->i_mtime);
  CXITIME_FROM_INODETIME(vattrP->va_ctime, inodeP->i_ctime);
  cache_copy_fh((void *)fhP, NFS_FH(inodeP));
  TRACE10(TRACE_PCACHE, 9, TRCID_CACHE_INODE2PATTR,
          "inode2pattr: inode 0x%X mode %o uid %d gid %d rdev %d inode %lld "
          "size %d mtime %ld ctime %ld fhsize %d",
          inodeP, inodeP->i_mode, inodeP->i_uid, inodeP->i_gid, inodeP->i_rdev,
          nfsiP->fileid, inodeP->i_size, inodeP->i_mtime.tv_sec,
          inodeP->i_ctime.tv_sec, fhP->size);
  TRACE8(TRACE_PCACHE, 9, TRCID_CACHE_INODE2PATTR_2,
         "inode2pattr: fh: %08x %08x %08x %08x %08x %08x %08x %08x",
         fh[0], fh[1], fh[2], fh[3], fh[4], fh[5], fh[6], fh[7]);
}

static inline struct dentry *
cache_invalidate(struct dentry *dP)
{
  int rc = 0;
  if ((rc = d_invalidate(dP)) == 0)
  {
    dput(dP);
    dP = NULL;
  }

  TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_INVALIDATE,
         "cache_invalidate exit: dP 0x%lX rc %d",
         dP, rc);
  return dP;
}

/* Revalidate a dentry that is found in the cache. Depending on the
   intent specified in the nameidata, this may cause the file to be
   opened at the server */
static inline struct dentry *
cache_revalidate(struct dentry *dP, struct nameidata *ndP)
{
  int rc = 0;

  if (dP->d_op && dP->d_op->d_revalidate)
  {
    rc = dP->d_op->d_revalidate(dP, ndP);
    if (rc > 0)
      goto out;
    if (rc == 0)
    {
      if (!d_invalidate(dP))
      {
        dput(dP);
        dP = NULL;
      }
    }
    else
    {
      dput(dP);
      dP = ERR_PTR(rc);
    }
  }
 out:
  TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_REVALIDATE,
         "cache_revalidate exit: dP 0x%lX rc %d",
         dP, rc);
  return dP;
}

/* Lookup a dentry given its name and the inode of the parent directory.
   Returns with a hold on the dentry.  This is similar to Linux's generic
   do_lookup() but that is not exported for use by external modules */
static struct dentry *
cache_lookup(struct dentry *ddP, const char *nameP, struct nameidata *ndP)
{
  struct dentry *dP = NULL;
  struct inode *dirP = ddP->d_inode;
  struct file *fileP = NULL;
  struct qstr name;
  int code = 0;

  /* If the name starts with a '/', we are probably looking up '.' */
  while (*nameP == '/')
    nameP++;

  if (!*nameP || (*nameP == '.' && !*(nameP+1)))
  {
    dP = cache_revalidate(dget(ddP), NULL);
    code = 1;
    goto out;
  }

  name.name = nameP;
  name.len = strlen(nameP);
  name.hash = full_name_hash(name.name, name.len);

  INODE_LOCK(dirP);

  /* Check if the dentry exists in the dcache */
  dP = d_lookup(ddP, &name);
  if (dP != NULL)
  {
    INODE_UNLOCK(dirP);
    TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_LOOKUP_1,
          "cache_lookup: found cached dentry 0x%lX for %s",
          dP, nameP);
    /* NFS only supports OPEN on regular files */
    if (dP->d_inode != NULL &&
        S_ISREG(dP->d_inode->i_mode) &&
        (ndP->flags & LOOKUP_OPEN))
    {
#ifdef LOOKUP_INTENT
      fileP = get_empty_filp();
      if (fileP == NULL)
      {
        dput(dP);
        dP = ERR_PTR(-ENFILE); /* FIX: ENOENT instead? */
        code = 2;
        goto out;
      }
      ndP->intent.open.file = fileP;
#endif
      dP = cache_revalidate(dP, ndP);
    }
    else
      dP = cache_revalidate(dP, NULL);
    if (dP == NULL)
      INODE_LOCK(dirP);
  }

  /* Check again for dP since revalidate may have invalidated it */
  if (dP == NULL)
  {
    struct dentry *newP = NULL;

#ifdef LOOKUP_INTENT
    /* Allocate a file struct for open */
    if (ndP->flags & LOOKUP_OPEN)
    {
      fileP = get_empty_filp();
      if (fileP == NULL)
      {
        dP = ERR_PTR(-ENFILE);  /* FIX: ENOENT instead? */
        code = 3;
        goto out;
      }
      ndP->intent.open.file = fileP;
    }
#endif
    newP = d_alloc(ddP, &name);
    if (newP == NULL)
    {
      dP = ERR_PTR(-ENOMEM);
      code = 4;
      goto out;
    }

    DBGASSERT(dirP->i_op->lookup != NULL);

    dP = dirP->i_op->lookup(dirP, newP, ndP);
    if (dP)
      dput(newP);
    else
      dP = newP;
    INODE_UNLOCK(dirP);
  }

  /* Is this a negative dentry? */
  if (!IS_ERR(dP) && dP->d_inode == NULL)
  {
    dput(dP);
    dP = ERR_PTR(-ENOENT);
    code = 5;
  }
 out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CACHE_LOOKUP,
         "cache_lookup exit: ddP 0x%lX name %X flags 0x%X ret dP 0x%lX code %d",
         ddP, nameP, ndP->flags, dP, code);
  return dP;
}

/**
 * @dentry: the directory in which to find a name
 * @name:   a pointer to a %NAME_MAX+1 char buffer to store the name
 * @child:  the dentry for the child directory.
 *
 * calls readdir on the parent until it finds an entry with
 * the same inode number as the child, and returns that.
 */
struct getdents_callback
{
  char *nameP;
  unsigned long ino;
  int found;
  int sequence;
};

static int filldir_one(void * __bufP, const char * nameP, int len,
                       loff_t pos, u64 ino, unsigned int d_type)
{
  struct getdents_callback *bufP = __bufP;
  int result = 0;

  bufP->sequence++;
  if (bufP->ino == ino) {
    memcpy(bufP->nameP, nameP, len);
    bufP->nameP[len] = '\0';
    bufP->found = 1;
    result = -1;
  }
  return result;
}

/* Do a readdir on the parent to find the matching name */
static int cache_get_name(struct vfsmount *mntP, struct dentry *dentryP,
                          char *nameP, struct dentry *childP)
{
  struct inode *dirP = dentryP->d_inode;
  int error = 0;
  struct file *fileP = NULL;
  struct getdents_callback buffer;

  error = -ENOTDIR;
  if (!dirP || !S_ISDIR(dirP->i_mode))
    goto out;
  error = -EINVAL;
  if (!dirP->i_fop)
    goto out;
  /*
   * Open the directory ...
   */
  fileP = DENTRY_OPEN(dget(dentryP), mntP, O_RDONLY);
  if (IS_ERR(fileP))
  {
    error = PTR_ERR(fileP);
    goto out;
  }

  if (!fileP->f_op->readdir)
  {
    error = -EINVAL;
    goto out_close;
  }

  buffer.nameP = nameP;
  buffer.ino = childP->d_inode->i_ino;
  buffer.found = 0;
  buffer.sequence = 0;
  while (1) {
    int old_seq = buffer.sequence;

    error = vfs_readdir(fileP, (filldir_t)filldir_one, &buffer);

    if (error < 0)
      break;

    error = 0;
    if (buffer.found)
      break;
    error = -ENOENT;
    if (old_seq == buffer.sequence)
      break;
  }

 out_close:
  fput(fileP);
 out:
  return error;
}

/* Find a dentry associated with the specified inode.
   Returns with a hold on the dentry
   - if parentP is given, use it to find the correct one
   - if nameP is given, use it to find the matching name
   On error, releases hold on inode */
static struct dentry *
cache_find_dentry(struct vfsmount *mntP, struct dentry *parentP,
                  struct inode *inodeP, const char *nameP)
{
  struct dentry *resP = NULL;
  struct dentry *dirP = NULL;
  int noprogress = 0;
  int err = 0;

  /* If name is specified, try to find the right dentry */
  if (nameP != NULL)
  {
    struct dentry *ldP;
    spin_lock(&dcache_lock);
    list_for_each_entry(ldP, &inodeP->i_dentry, d_alias)
    {
      /* Check if the name matches */
      if (nameP && strcmp(ldP->d_name.name, nameP) == 0)
      {
        resP = dget_locked(ldP);
        break;
      }
    }
    spin_unlock(&dcache_lock);
    if (resP == NULL) /* didn't find it */
    {
      if (parentP != NULL)
      {
        struct qstr q;

        q.name = nameP;
        q.len = strlen(nameP);
        q.hash = full_name_hash(q.name, q.len);
        resP = d_alloc(parentP, &q);
        if (resP != NULL)
          d_add(resP, inodeP);
      }
      else
        resP = d_alloc_anon(inodeP);
    }
  }
  else
    resP = d_alloc_anon(inodeP);

  /* Return error if we couldn't get a valid dentry;
     need to release hold on inode */
  if (resP == NULL)
  {
    resP = ERR_PTR(-ENOENT);
    goto out_iput;
  }
  else if (IS_ERR(resP))
    goto out;

  if (S_ISDIR(resP->d_inode->i_mode) && (resP->d_flags & DCACHE_DISCONNECTED))
  {
    /* the directory is not connected, we must connect it */
    ;
  }
  else
    goto out;

  dirP = dget(resP);

  /* Note: The following code is similar in concept to find_exported_dentry
     used by exportfs */

  /* Now we need to make sure that dirP is properly connected.
   * It may already be, as the flag isn't always updated when connection
   * happens.
   * So, we walk up parent links until we find a connected directory,
   * or we run out of directories.  Then we find the parent, find
   * the name of the child in that parent, and do a lookup.
   * This should connect the child into the parent
   * We then repeat.
   */
  while ((dirP->d_flags & DCACHE_DISCONNECTED) && noprogress++ < 10)
  {
    struct dentry *pdP = dirP;

    dget(pdP);
    spin_lock(&pdP->d_lock);
    while (!IS_ROOT(pdP) && (pdP->d_parent->d_flags & DCACHE_DISCONNECTED))
    {
      struct dentry *pddP = pdP->d_parent;
      dget(pddP);
      spin_unlock(&pdP->d_lock);
      dput(pdP);
      pdP = pddP;
      spin_lock(&pdP->d_lock);
    }
    spin_unlock(&pdP->d_lock);

    if (!IS_ROOT(pdP))
    {
      /* found a connected parent */
      spin_lock(&pdP->d_lock);
      pdP->d_flags &= ~DCACHE_DISCONNECTED;
      spin_unlock(&pdP->d_lock);
      noprogress = 0;
    }
    else
    {
      /* Note: This code is intentionally duplicated pending the fix below */
      /* FIX: We really need to check here if we are at the root of the NFS
       * mount with something like (pdP == mntP->mnt_sb->s_root) but with
       * kernels > 2.6.18, the NFS superblock is shared by all mounts to a server
       * so the condition may not work
       */
      spin_lock(&pdP->d_lock);
      pdP->d_flags &= ~DCACHE_DISCONNECTED;
      spin_unlock(&pdP->d_lock);
      noprogress = 0;
    }
    dput(pdP);
  }

  if (dirP->d_flags & DCACHE_DISCONNECTED)
  {
    /* something went wrong - oh-well */
    dput(resP);
    resP = ERR_PTR(-ESTALE);
    goto out_dput;
  }

  /* if we weren't after a directory, have one more step to go */
  if (resP != dirP)
  {
    char *buf;
    struct dentry *nresP;

    buf = (char *) __get_free_page(GFP_USER);
    if (buf == NULL)
    {
      dput(resP);
      resP = ERR_PTR(-ENOMEM);
      goto out_dput;
    }

    err = cache_get_name(mntP, dirP, buf, resP);
    if (!err)
    {
      INODE_LOCK(dirP->d_inode);
      nresP = lookup_one_len(buf, dirP, strlen(buf));
      INODE_UNLOCK(dirP->d_inode);
      if (!IS_ERR(nresP))
      {
        if (nresP->d_inode)
        {
          dput(resP);
          resP = nresP;
        }
        else
          dput(nresP);
      }
    }
    free_page((unsigned long) buf);
  }

 out_dput:
  dput(dirP);
 out_iput:
  if (IS_ERR(resP))
    iput(inodeP);
 out:
  TRACE3(TRACE_PCACHE, 9, TRCID_CACHE_FIND_DENTRY_EXIT,
         "cache_find_dentry: inode %lld name %s ret dentry 0x%lX",
         inodeP->i_ino, nameP, resP);
  return resP;
}

struct nfs_find_desc
{
  cxiIno64_t ino;
  struct nfs_fh *fhP;
};
  
static int
nfs_find_inode(struct inode *inodeP, void *opaqueP)
{
  struct nfs_find_desc *descP = (struct nfs_find_desc *)opaqueP;
  struct nfs_fh *fhP = descP->fhP;
  cxiIno64_t ino = descP->ino;
  
  if (inodeP->i_ino != ino ||
      cache_compare_fh(NFS_FH(inodeP), fhP) ||
      is_bad_inode(inodeP) ||
      NFS_STALE(inodeP))
    return 0;
  return 1;
}
  
/* Get NFS inode from filehandle and attributes
   Returns with a hold on the inode */
static struct inode *
nfs_iget(void *vfsP, cxiPattr_t *pattrP)
{
  struct vfsmount *mntP = (struct vfsmount *)vfsP;
  struct nfs_fh *fhP = (struct nfs_fh *)&pattrP->pa_fh;
  struct nfs_fattr fattr;
  struct inode *inodeP = ERR_PTR(-ENOENT);
  __u32 *fh = (__u32 *)fhP->data;
  struct nfs_find_desc desc = {
    .fhP = fhP,
    .ino = pattrP->pa_vattr.va_ino
  };

  if (fhP->size == 0)
    goto out;

  /* We don't want to call fhget() directly here before checking if the inode
     is in the cache. If we do, a cached inode with (valid) attributes will
     be updated with the ones we are passing to fhget(). We only want to use
     the local cached attributes to help construct a new inode. So we check first
     if there is a cached inode. If we find one, we return it directly - it will
     be revalidated later */
  inodeP = ilookup5(mntP->mnt_sb, desc.ino, nfs_find_inode, &desc);
  if (inodeP == NULL)
  {
    vattr2fattr(mntP->mnt_sb, &pattrP->pa_vattr, &fattr);
    inodeP = fhget(mntP->mnt_sb, fhP, &fattr);
  }

 out:
  TRACE3(TRACE_PCACHE, 4, TRCID_NFS_IGET,
         "nfs_iget: inode %lld, fh size %d, inodeP 0x%lX",
         pattrP->pa_vattr.va_ino, fhP->size, inodeP);
  TRACE8(TRACE_PCACHE, 9, TRCID_NFS_IGET_FH,
         "nfs_iget: fh: %08x %08x %08x %08x %08x %08x %08x %08x",
         fh[0], fh[1], fh[2], fh[3], fh[4], fh[5], fh[6], fh[7]);
  return inodeP;
}

/* Get NFS dentry from filehandle and attributes.
   name (and parent dentry) is used to find the matching dentry
   Returns with a hold on the dentry */
static struct dentry *
nfs_dget_name(void *vfsP, struct dentry *parentP,
              const char *nameP, cxiPattr_t *pattrP)
{
  struct inode *inodeP = NULL;
  struct dentry *dentryP = ERR_PTR(-ENOENT);

  inodeP = nfs_iget(vfsP, pattrP);
  if (!IS_ERR(inodeP))
  {
    dentryP = cache_find_dentry(vfsP, parentP, inodeP, nameP);
    if (!IS_ERR(dentryP))
    {
      struct kstat stat;
      int rc = 0;
      /* We need to refresh the inode since we used a (possibly stale) fh
         to construct the dentry from it */
      rc = vfs_getattr(vfsP, dentryP, &stat);
      TRACE3(TRACE_PCACHE, 4, TRCID_NFS_GETATTR,
             "nfs_dget_name: vfs_getattr inode 0x%lX name %s ret %d",
             inodeP, dentryP->d_name.name, rc);
      if (rc != 0)
      {
        dput(dentryP);
        dentryP = ERR_PTR(rc);
      }
    }
  }
  else
    dentryP = (struct dentry *)inodeP; /* err */

  return dentryP;
}

/* Get NFS dentry from filehandle and attributes.
   Returns with a hold on the dentry */
static inline struct dentry *
nfs_dget(void *vfsP, cxiPattr_t *pattrP)
{
  return nfs_dget_name(vfsP, NULL, NULL, pattrP);
}

/*
 * Cache operations
 */

/* Initialization routine called at mount time to establish link
   between a fileset and NFS vfsmount. Returns with a hold on
   the NFS mount so it doesn't disappear under us */
int cxiCacheInit(const char *sgName, UInt32 filesetId, void **vfsPP, void **filePP)
{
  struct nameidata nd;
  char *pageP = NULL;
  cxiPattr_t *pattrP = NULL;
  int rc = -ENODEV, code = 0;

  if (sgName == NULL || *sgName == '\0')
  {
    code = 1;
    goto xerror;
  }

  pageP = (char *) __get_free_page(GFP_USER);
  if (pageP == NULL)
  {
    code = 2;
    goto xerror;
  }

  sprintf(pageP, "/var/mmfs/cache/%s-%d", sgName, filesetId);
  rc = path_lookup(pageP, LOOKUP_FOLLOW | LOOKUP_DIRECTORY, &nd);
  if (rc != 0)
  {
    code = 3;
    goto xerror;
  }

  /* Put a hold count on the vfsmount */
  *vfsPP = mntget(ND_PATH(nd).mnt);

  if (filePP != NULL)
  {
    /* Open the special file on home server to read xattr.
       Ignore errors because the file may not be present */
    *filePP = NULL;
    pattrP = (cxiPattr_t *)pageP;
    rc = cxiCacheLookupRoot(*vfsPP, pattrP);
    if (rc == 0)
    {
      void *fP = NULL;
      loff_t fileSize = 0;
      rc = cxiCacheOpen(*vfsPP, pattrP, PCACHE_XATTR_CTL_FILE, O_RDWR|O_DIRECT,
                        NULL, &fP, &fileSize);
      if (rc == 0)
        *filePP = fP;
    }
  }

  PATH_PUT(&nd);
  rc = 0;

 xerror:
  free_page((unsigned long) pageP);
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_INIT,
         "cxiCacheInit: fileset id %d remote vfsP 0x%lX fileP 0x%lX rc %d code %d",
         filesetId, *vfsPP, (filePP != NULL) ? *filePP : 0, rc, code);
  return -rc;
}

/* Exit routine called at umount time to release the hold on
   the NFS mount established at GPFS mount time */
int cxiCacheExit(void *vfsP, void *fileP)
{
  DBGASSERT(vfsP != NULL);

  if (fileP != NULL)
    cxiCacheClose(fileP, NULL);
  mntput(vfsP);
  TRACE2(TRACE_PCACHE, 4, TRCID_CXICACHE_EXIT,
         "cxiCacheExit: nfs vfsP 0x%lX xattr fileP 0x%lX", vfsP, fileP);
  return 0;
}

/* Return NFS fh and attrs of root */
int
cxiCacheLookupRoot(void *vfsP, cxiPattr_t *pattrP)
{
  struct dentry *rootP = ((struct vfsmount *)vfsP)->mnt_root;
  int rc = -ENOENT;

  if (rootP && rootP->d_inode)
  {
    inode2pattr(rootP->d_inode, pattrP);
    rc = 0;
  }
  TRACE2(TRACE_PCACHE, 4, TRCID_CXICACHE_LOOKUP_ROOT,
         "cxiCacheLookupRoot: returned ino %lld rc %d",
         pattrP->pa_vattr.va_ino, rc);
  return -rc;
}

/* Do lookup on a name, given the parent fh and attrs,
   and if successful, return resulting fh and attrs */
int
cxiCacheLookup(void *vfsP, cxiPattr_t *pattrP, char *nameP, cxiPattr_t *cpattrP)
{
  struct dentry *ddP = NULL;
  struct dentry *dP = NULL;
  struct nameidata nd = {};
  int rc = 0;

  /* Get dentry of parent */
  if (*nameP == '/')
    ddP = dget(((struct vfsmount *)vfsP)->mnt_root);
  else
    ddP = nfs_dget(vfsP, pattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    goto out;
  }

  /* We already have a hold on ddP */
  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP;
  dP = cache_lookup(ddP, nameP, &nd);
  if (IS_ERR(dP))
  {
    rc = PTR_ERR(dP);
    goto out_dput;
  }

  inode2pattr(dP->d_inode, cpattrP);
  dput(dP);

 out_dput:
  PATH_PUT(&nd);
 out:
  TRACE4(TRACE_PCACHE, 4, TRCID_CXICACHE_LOOKUP,
         "cxiCacheLookup: name %s, dir %d, returned ino %lld rc %d",
         nameP, pattrP->pa_vattr.va_ino, cpattrP->pa_vattr.va_ino, rc);
  return -rc;
}

/* Open a file identified by name, fh and attrs (including parent's) */
int
cxiCacheOpen(void *vfsP, cxiPattr_t *parentPattrP, char *nameP, int flags,
             cxiPattr_t *pattrP, void **filePP, loff_t *fileSizeP)
{
  struct dentry *ddP = NULL, *dP = NULL;
  struct nameidata nd;
  struct file *fileP = NULL;
  int namei_flags = 0, lookup_flags = 0, rc = 0;

  TRACE2(TRACE_PCACHE, 4, TRCID_CXICACHE_OPEN_IN,
         "cxiCacheOpen: name %s flags 0%o", nameP, flags);

  /* Init return values */
  *filePP = NULL;
  *fileSizeP = 0;

  if (*nameP == '\0')
  {
    dP = nfs_dget(vfsP, parentPattrP);
    if (IS_ERR(dP))
    {
      rc = PTR_ERR(dP);
      goto out;
    }
    DBGASSERT(S_ISDIR(dP->d_inode->i_mode));
    ddP = dget(dP->d_parent);
    nameP = (char *)dP->d_name.name;
  }
  else
  {
    /* Lookup up the dentries from the filehandle/attributes */
    ddP = nfs_dget(vfsP, parentPattrP);
    if (IS_ERR(ddP))
    {
      rc = PTR_ERR(ddP);
      goto out;
    }
  }

  /* We already have a hold on ddP */
  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP;

  /* We always want to support large files */
  flags |= O_LARGEFILE;

  /* Map open flags to open intent flags */
  namei_flags = flags;
  if ((namei_flags+1) & O_ACCMODE)
    namei_flags++;

  lookup_flags = LOOKUP_FOLLOW;
  if (flags & O_NOFOLLOW)
    lookup_flags &= ~LOOKUP_FOLLOW;
  if (flags & O_DIRECTORY)
    lookup_flags |= LOOKUP_DIRECTORY;

  nd.flags = lookup_flags | LOOKUP_OPEN;
  INTENT_OPEN_FLAGS(nd.intent) = namei_flags;
  INTENT_OPEN_CREATE_MODE(nd.intent) = 0;

#ifdef LOOKUP_INTENT
  /* Note: file struct below will be allocated by cache_lookup, because
     we do not want to do a open revalidate for directories */
  nd.intent.open.file = NULL;
#endif

  if (dP == NULL)
  {
    dP = cache_lookup(ddP, nameP, &nd);
    if (IS_ERR(dP))
    {
      rc = PTR_ERR(dP);
      goto out_release;
    }
  }

  /* Initialize fileP if it hasn't already been done */
#ifdef LOOKUP_INTENT
  fileP = nd.intent.open.file;
#endif
  if (fileP == NULL || fileP->f_dentry == NULL)
  {
    /* We would like to call __dentry_open() here but that's not exported.
       So we can't reuse the fileP we allocated earlier and must create
       a new one. We can't even free the one we have because it's not
       initialized properly so fput() would fail and fput_filp() is not
       exported either! Ugh... */
    if (fileP != NULL)
    {
      TRACE2(TRACE_PCACHE, 1, TRCID_CXICACHE_OPEN_FPUT,
             "cxiCacheOpen: BUG! filp 0x%lX for name %s will not be freed",
             fileP, nameP);
      //printk("cxiCacheOpen: BUG! fileP for name %s will not be freed\n", nameP);
      //fput(fileP);
    }
    fileP = DENTRY_OPEN(dget(dP), mntget(vfsP), flags);
    if (IS_ERR(fileP))
    {
      rc = PTR_ERR(fileP);
      goto out_dput;
    }
  }

  *filePP = fileP;
  *fileSizeP = dP->d_inode->i_size;
  if (pattrP != NULL)
    inode2pattr(dP->d_inode, pattrP);

 out_dput:
  dput(dP);
 out_release:
  PATH_PUT(&nd);
 out:
  TRACE7(TRACE_PCACHE, 4, TRCID_CXICACHE_OPEN,
         "cxiCacheOpen: parent ino %d name %s flags 0x%X ret ino %d fileP 0x%lX size %d rc %d",
         parentPattrP->pa_vattr.va_ino, nameP, flags,
         (rc == 0 && dP && dP->d_inode) ? dP->d_inode->i_ino : -1, fileP, *fileSizeP, rc);
  return -rc;
}

/* Create a new object and return its fh and attrs */
int
cxiCacheCreate(void *vfsP, cxiPattr_t *parentPattrP, char *nameP, int mode,
               uid_t uid, gid_t gid, cxiDev_t rdev, cxiPattr_t *cpattrP)
{
  struct dentry *ddP = NULL;
  struct dentry *dP = NULL;
  struct qstr name;
  struct nameidata nd = {};
  dev_t dev;
  int code = 0, rc = 0;

  /* Get dentry of parent */
  ddP = nfs_dget(vfsP, parentPattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    code = 1;
    goto out;
  }

  name.name = nameP;
  name.len = strlen(nameP);
  name.hash = full_name_hash(name.name, name.len);

  dP = d_lookup(ddP, &name);
  if (dP == NULL)
  {
    dP = d_alloc(ddP, &name);
    if (dP == NULL)
    {
      rc = -ENOMEM;
      code = 2;
      goto out;
    }
  }
  else if (dP->d_inode != NULL)
  {
    rc = -EEXIST;
    if (cpattrP)
      inode2pattr(dP->d_inode, cpattrP);
    code = 3;
    goto out;
  }

  switch (mode & S_IFMT)
  {
    case S_IFREG:
      /* Need to pass in a valid nameidata since NFS looks at the flags */
      ND_PATH(nd).mnt = mntget(vfsP);
      ND_PATH(nd).dentry = dget(ddP);
      nd.flags = LOOKUP_CREATE;
      INTENT_OPEN_FLAGS(nd.intent) = FMODE_READ | FMODE_WRITE | O_CREAT;
      INTENT_OPEN_CREATE_MODE(nd.intent) = mode;
      rc = vfs_create(ddP->d_inode, dP, mode, &nd);
      PATH_PUT(&nd);
      break;

    case S_IFDIR:
      rc = VFS_MKDIR(vfsP, ddP->d_inode, dP, mode);
      break;

    case S_IFCHR:
    case S_IFBLK:
    case S_IFIFO:
    case S_IFSOCK:
      dev = ((rdev>>16)<<20)|(rdev & 0xFFFF);
      rc = VFS_MKNOD(vfsP, ddP->d_inode, dP, mode, dev);
      break;

    default:
      rc = -EINVAL;
  }

  if (rc == 0)
  {
    if (uid != 0 || gid != 0)
    {
      /* FIX: This check needs to be updated when we create as non-root user.
         If we run in the context of the caller, then no attribute update is required.
         Until then, if a non-root user performed the original create, we need
         to update remote uid/gid */
      struct iattr iattr;
      iattr.ia_valid = ATTR_UID|ATTR_GID;
      iattr.ia_uid = uid;
      iattr.ia_gid = gid;
      rc = dP->d_inode->i_op->setattr(dP, &iattr);
    }
    if (cpattrP)
      inode2pattr(dP->d_inode, cpattrP);
  }

 out:
  if (!IS_ERR(ddP))
    dput(ddP);
  if (dP && !IS_ERR(dP))
    dput(dP);

  TRACE7(TRACE_PCACHE, 4, TRCID_CXICACHE_CREATE,
         "cxiCacheCreate: name %s mode 0%o dir %ld uid %d gid %d rc %d code %d",
         nameP, mode, parentPattrP->pa_vattr.va_ino, uid, gid, rc, code);
  return -rc;
}

/* Remove an object, given its name and parent's fh and attrs */
int
cxiCacheRemove(void *vfsP, cxiPattr_t *parentPattrP, char *nameP, int type)
{
  struct dentry *ddP = NULL;
  struct dentry *dP = NULL;
  struct nameidata nd = {};
  int code = 0, rc = 0;

  /* Get dentry of parent */
  ddP = nfs_dget(vfsP, parentPattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    code = 1;
    goto out;
  }

  /* We already have a hold on ddP */
  ND_PATH(nd).mnt = mntget(vfsP);
  ND_PATH(nd).dentry = ddP;
  dP = cache_lookup(ddP, nameP, &nd);
  if (IS_ERR(dP))
  {
    rc = PTR_ERR(dP);
    code = 2;
    goto out_dput;
  }

  switch (type)
  {
    case S_IFREG:
    case S_IFCHR:
    case S_IFBLK:
    case S_IFIFO:
    case S_IFSOCK:
      rc = VFS_UNLINK(vfsP, ddP->d_inode, dP);
      break;

    case S_IFDIR:
      rc = VFS_RMDIR(vfsP, ddP->d_inode, dP);
      break;

    default:
      code = 3;
      rc = -EINVAL;
  }
  dput(dP);

 out_dput:
  PATH_PUT(&nd);

 out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_REMOVE,
         "cxiCacheRemove: name %s type 0x%X dir %d, rc %d code %d",
         nameP, type, parentPattrP->pa_vattr.va_ino, rc, code);
  return -rc;
}

/* Create a hard link to an object identified by fh and attrs */
int
cxiCacheLink(void *vfsP, cxiPattr_t *parentPattrP, cxiPattr_t *pattrP, char *nameP)
{
  struct dentry *ddP = NULL, *sdP = NULL, *tdP = NULL;
  struct qstr name;
  int code = 0, rc = 0;

  /* Get dentry of parent and source */
  ddP = nfs_dget(vfsP, parentPattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    code = 1;
    goto out;
  }

  /* We don't have the source name, so can't call nfs_dget_name */
  sdP = nfs_dget(vfsP, pattrP);
  if (IS_ERR(sdP))
  {
    rc = PTR_ERR(sdP);
    code = 2;
    goto out_ddput;
  }

  name.name = nameP;
  name.len = strlen(nameP);
  name.hash = full_name_hash(name.name, name.len);

  tdP = d_lookup(ddP, &name);
  if (tdP == NULL)
  {
    tdP = d_alloc(ddP, &name);
    if (tdP == NULL)
    {
      rc = -ENOMEM;
      code = 3;
      goto out_sdput;
    }
  }

  rc = VFS_LINK(vfsP, sdP, ddP->d_inode, tdP);
  dput(tdP);

 out_sdput:
  dput(sdP);
 out_ddput:
  dput(ddP);
 out:
  TRACE4(TRACE_PCACHE, 4, TRCID_CXICACHE_LINK,
         "cxiCacheLink: serialno %ld, new name %s, rc %d code %d",
         pattrP->pa_vattr.va_ino, nameP, rc, code);
  return -rc;
}

/* Create a symlink to an object identified by name and parent's fh and attrs */
int
cxiCacheSymlink(void *vfsP, cxiPattr_t *parentPattrP,
                char *targetNameP, char *newNameP, cxiPattr_t *cpattrP)
{
  struct dentry *ddP = NULL;
  struct dentry *dP = NULL;
  struct qstr name;
  int code = 0, rc = 0;

  /* Get dentry of parent */
  ddP = nfs_dget(vfsP, parentPattrP);
  if (IS_ERR(ddP))
  {
    rc = PTR_ERR(ddP);
    code = 1;
    goto out;
  }

  name.name = newNameP;
  name.len = strlen(newNameP);
  name.hash = full_name_hash(name.name, name.len);

  dP = d_lookup(ddP, &name);
  if (dP == NULL)
  {
    dP = d_alloc(ddP, &name);
    if (dP == NULL)
    {
      rc = -ENOMEM;
      code = 2;
      goto out_dput;
    }
  }

  rc = VFS_SYMLINK(vfsP, ddP->d_inode, dP, targetNameP);
  if (rc == 0 && cpattrP != NULL)
    inode2pattr(dP->d_inode, cpattrP);

  dput(dP);

 out_dput:
  dput(ddP);
 out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_SYMLINK,
         "cxiCacheSymlink: dir %d target name %s new name %s, rc %d code %d",
         parentPattrP->pa_vattr.va_ino, targetNameP, newNameP, rc, code);
  return -rc;
}

/* Get attributes on an object identified by fh and attrs
   Use a flag to optionally force an update of attributes from the
   remove server */
int
cxiCacheGetattr(void *vfsP, cxiPattr_t *pattrP, Boolean forceFlag)
{
  struct inode *iP = NULL;
  int rc = 0;
  struct kstat stat;
  struct dentry dentry;

  /* Get inode of object */
  iP = nfs_iget(vfsP, pattrP);
  if (IS_ERR(iP))
  {
    rc = PTR_ERR(iP);
    goto out;
  }

  dentry.d_inode = iP;

  /* Force attr invalid to refresh from server */
  if (forceFlag)
    NFS_CACHE_VALIDITY_FLAGS(iP) |= NFS_INO_INVALID_ATTR;
  rc = vfs_getattr(vfsP, &dentry, &stat);
  if (rc == 0)
    inode2pattr(iP, pattrP);

  iput(iP);

 out:
  TRACE2(TRACE_VNODE, 4, TRCID_CACHE_GETATTR,
         "cxiCacheGetattr: ino %d rc %d",
         pattrP->pa_vattr.va_ino, rc);
  return -rc;
}

/* Set attributes on an object identified by fh and attrs.
 * Refresh pattrP with latest attributes. */
int
cxiCacheSetattr(void *vfsP, cxiPattr_t *pattrP, int attrset, long arg1, long arg2)
{
  struct dentry *dP = NULL;
  struct iattr iattr;
  int rc = 0;

  /* Get inode of object */
  dP = nfs_dget(vfsP, pattrP);
  if (IS_ERR(dP))
  {
    rc = PTR_ERR(dP);
    goto out;
  }

  iattr.ia_valid = attrset;
  if (attrset == (ATTR_MODE | ATTR_SIZE | ATTR_UID | ATTR_GID))
  {
    /* called during recovery - pick attributes from attr */
    iattr.ia_mode = pattrP->pa_vattr.va_mode;
    iattr.ia_size = pattrP->pa_vattr.va_size;
    iattr.ia_uid = pattrP->pa_vattr.va_uid;
    iattr.ia_gid = pattrP->pa_vattr.va_gid;
  }
  else if (attrset & ATTR_MODE)
    iattr.ia_mode = (umode_t)arg1;
  else if (attrset & ATTR_SIZE)
    iattr.ia_size = (loff_t)arg1;
  else if (attrset & (ATTR_UID|ATTR_GID))
  {
    if (attrset & ATTR_UID)
      iattr.ia_uid = (uid_t)arg1;
    if (iattr.ia_valid & ATTR_GID)
      iattr.ia_gid = (gid_t)arg2;
  }

  /* FIX: we really need to call notify_change() here but it requires a
     valid dentry so dnotify_parent() doesn't oops the kernel. For now,
     it appears sufficient to just call setattr instead */
  rc = dP->d_inode->i_op->setattr(dP, &iattr);
  if (rc == 0)
  {
    /* Update attributes that _may_ be returned on a successful setattr (nfsv4) */
    inode2pattr(dP->d_inode, pattrP);
  }

  dput(dP);

 out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_SETATTR,
         "cxiCacheSetattr: ino %d attrset %d args 0x%lX 0x%lX rc %d",
         pattrP->pa_vattr.va_ino, attrset, arg1, arg2, rc);
  return -rc;
}

/* Close a file */
int
cxiCacheClose(void *fP, cxiPattr_t *pattrP)
{
  struct file *fileP = (struct file *)fP;
  struct inode *inodeP = fileP->f_dentry->d_inode;
  int rc = 0;

  if (!igrab(inodeP))
    DBGASSERT(0);

  rc = filp_close(fileP, current->files);

  if (pattrP != NULL)
    inode2pattr(inodeP, pattrP);

  iput(inodeP);

  TRACE3(TRACE_PCACHE, 4, TRCID_CXICACHE_CLOSE,
         "cxiCacheClose: fP 0x%lX name %s, rc %d",
         fileP, (fileP->f_dentry) ? (char *)fileP->f_dentry->d_name.name : "null", rc);
  return -rc;
}

/* Perform READDIR operation on a dir */
int
cxiCacheReaddir(void *fP, void *fnP, void *argP)
{
  struct file *fileP = (struct file *)fP;
  int rc;

  rc = vfs_readdir(fileP, fnP, argP);
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_READDIR,
         "cxiCacheReaddir: fP 0x%lX name %s flags 0%o mode %d rc %d",
         fileP, fileP->f_dentry->d_name.name, fileP->f_flags, fileP->f_mode, rc);
  return -rc;
}

/* Read a file (already open) starting at offsetP */
int cxiCacheRead(void *fP, char *bufP, size_t bufSize,
                 offset_t *offsetP, ssize_t *bytesRead, Boolean isUserBuf)
{
  struct file *fileP = (struct file *)fP;
  ssize_t rc = 0;
  int err = 0;

  DBGASSERT(fileP != NULL);
  *bytesRead = 0;
  if (isUserBuf)
    rc = vfs_read(fileP, (char __user *)bufP, bufSize, offsetP);
  else
    rc = kernel_read(fileP, *offsetP, bufP, bufSize);
  if (rc > 0)
  {
    *offsetP += rc;
    *bytesRead = rc;
    fileP->f_pos = *offsetP;
  }
  else
    err = -rc;

  TRACE7(TRACE_PCACHE, 4, TRCID_CXICACHE_READ,
         "cxiCacheRead: fileP 0x%lX flags 0%o mode %d offset %d desired %d read %d rc %d",
         fileP, fileP->f_flags,
         fileP->f_mode, *offsetP, bufSize, *bytesRead, err);
  return err;
}

/* Write to a file (already opened) starting at offsetP
   Return updated attributes */
int cxiCacheWrite(void *fP, char *bufP, size_t bufSize, offset_t *offsetP,
                  ssize_t *bytesWritten, Boolean isUserBuf)
{
  struct file *fileP = (struct file *)fP;
  mm_segment_t old_fs;
  loff_t pos = *offsetP;
  ssize_t rc = 0;
  int err = 0;

  DBGASSERT(fP != NULL);
  if (isUserBuf)
    rc = vfs_write(fileP, (void __user *)bufP, bufSize, &pos);
  else
  {
    old_fs = get_fs();
    set_fs(get_ds());
    /* The cast to a user pointer is valid due to the set_fs() */
    rc = vfs_write(fileP, (void __user *)bufP, bufSize, &pos);
    set_fs(old_fs);
  }

  *bytesWritten = 0;
  if (rc > 0)
  {
    *offsetP += rc;
    *bytesWritten = rc;
    fileP->f_pos = pos;
  }
  else
    err = -rc;

  TRACE7(TRACE_PCACHE, 4, TRCID_CACHE_CXIWRITE,
         "cxiCacheWrite: fileP 0x%lX flags 0%o mode %d offset %d desired %d written %d rc %d",
         fileP, fileP->f_flags,
         fileP->f_mode, *offsetP, bufSize, *bytesWritten, err);
  return err;
}

/* Rename a file, given names, fh's and attrs */
int
cxiCacheRename(void *vfsP, cxiPattr_t *sDirPattrP, cxiPattr_t *sPattrP, char *sNameP,
               cxiPattr_t *tDirPattrP, cxiPattr_t *tPattrP, char *tNameP)
{
  struct dentry *sdP = NULL, *sddP = NULL, *tdP = NULL, *tddP = NULL;
  int rc = 0, code = 0;

  sddP = nfs_dget(vfsP, sDirPattrP);
  if (IS_ERR(sddP))
  {
    rc = PTR_ERR(sddP);
    code = 1;
    goto out;
  }

  if (tDirPattrP != NULL)
  {
    tddP = nfs_dget(vfsP, tDirPattrP);
    if (IS_ERR(tddP))
    {
      rc = PTR_ERR(tddP);
      code = 2;
      goto out_sddput;
    }
  }
  else
    tddP = sddP;

  sdP = nfs_dget_name(vfsP, sddP, sNameP, sPattrP);
  if (IS_ERR(sdP))
  {
    rc = PTR_ERR(sdP);
    code = 3;
    goto out_tddput;
  }

  /* Sanity check */
  if (sddP != sdP->d_parent)
  {
    rc = -EINVAL;
    code = 5;
    goto out_sdput;
  }

  if (tPattrP != NULL)
  {
    tdP = nfs_dget_name(vfsP, tddP, tNameP, tPattrP);
    if (IS_ERR(tdP))
    {
      rc = PTR_ERR(tdP);
      code = 4;
      goto out_sdput;
    }
  }
  else
  {
    struct qstr tname;
    tname.name = tNameP;
    tname.len = strlen(tNameP);
    tname.hash = full_name_hash(tname.name, tname.len);

    tdP = d_lookup(tddP, &tname);
    if (tdP == NULL)
      tdP = d_alloc(tddP, &tname);
    if (IS_ERR(tdP))
    {
      code = 6;
      rc = PTR_ERR(tdP);
      goto out_sdput;
    }
  }

  rc = VFS_RENAME(vfsP, sddP->d_inode, sdP, tddP->d_inode, tdP);

  dput(tdP);

 out_sdput:
/* XXX: Is this the right place to invalidate? */
/*  dput(sdP); */
  cache_invalidate(sdP);
out_tddput:
  if (tDirPattrP != NULL)
    dput(tddP);
 out_sddput:
  dput(sddP);
 out:
  TRACE5(TRACE_PCACHE, 4, TRCID_CXICACHE_RENAME,
         "cxiCacheRename: source name %s target name %s source dir %d, rc %d code %d",
         sNameP, tNameP, sDirPattrP->pa_vattr.va_ino, rc, code);
  return -rc;
}

/* Get parent inode and file name used at open time from struct file */
int cxiCacheOpenState(void *dP, cxiNode_t **cnP, char **nameP)
{
  struct inode *inodeP = NULL;
  struct dentry *dentryP = dP;
  int rc = -ENOENT;

  if (dentryP == NULL)
    goto out;

  *nameP = (char *)dentryP->d_name.name;
  if (dentryP->d_parent == NULL)
    goto out;

  inodeP = (void *)dentryP->d_parent->d_inode;
  if (inodeP == NULL)
    goto out;

  *cnP = VP_TO_CNP(inodeP);
  rc = 0;

out:
  TRACE2(TRACE_PCACHE, 4, TRCID_CXICACHE_OPEN_EXIT,
         "cxiCacheOpenState: dentryP 0x%lX rc %d", dentryP, rc);
  return -rc;
}

/* Get the address of an unexported function - nfs_fhget */
void
cxiCacheSetaddr(void *addr)
{
  fhget = addr;
}

int
cxiCacheReadlink(void *vfsP, cxiPattr_t *pattrP, char *targetP)
{
  struct dentry *dentryP = NULL;
  struct nameidata nd = {};
  mm_segment_t oldfs;
  char *linkP = (char *) __get_free_page(GFP_USER);
  int len = 0, rc = 0;

  if (linkP == NULL)
  {
    rc = -ENOMEM;
    goto out;
  }

  dentryP = nfs_dget(vfsP, pattrP);
  if (IS_ERR(dentryP))
  {
    rc = PTR_ERR(dentryP);
    goto out_free;
  }

  oldfs = get_fs(); set_fs(KERNEL_DS);
  len = dentryP->d_inode->i_op->readlink(dentryP, linkP, PAGE_SIZE);
  set_fs(oldfs);

  if (len < 0)
  {
    rc = len;
    goto out_dput;
  }

  if (len > CXI_PATH_MAX)
  {
    /* FIX: we don't support symlinks > CXI_PATH_MAX */
    rc = -ENAMETOOLONG;
    goto out_dput;
  }

  strncpy(targetP, linkP, len);
  targetP[len] = '\0';

  rc = 0;

 out_dput:
  dput(dentryP);
 out_free:
  free_page((unsigned long)linkP);
 out:
  TRACE3(TRACE_PCACHE, 4, TRCID_CXICACHE_READLINK,
         "cxiCacheReadLink: targetP %s rc %d len %d", targetP, rc, len);
  return -rc;
}

void
cxiDeleteDCacheEntry(void *dP)
{
  struct dentry *dentryP = (struct dentry *)dP;
  DBGASSERT(dentryP != NULL);

  TRACE3(TRACE_VNODE, 4, TRCID_DEL_DCACHE_ENTRY,
         "cxiDeleteDCacheEntry: dentry 0x%lX d_inode 0x%lX (name '%s')\n",
         dentryP, dentryP->d_inode, dentryP->d_name.name);

  d_delete(dentryP);
}

int
cxiCacheNfsOpen(const char *sgNameP, UInt32 filesetId, cxiPattr_t *pattrP,
                char *nameP, int flags, int *fdP)
{
  void *vfsP = NULL;
  struct file *fileP = NULL;
  loff_t fileSize = 0;
  int fd = get_unused_fd();
  int rc = 0, code = 0;

  if (fd < 0)
  {
    rc = fd;
    code = 1;
    goto out;
  }

  rc = cxiCacheInit(sgNameP, filesetId, &vfsP, NULL);
  if (rc != 0)
  {
    code = 2;
    goto out;
  }

  rc = cxiCacheOpen(vfsP, pattrP, nameP, flags, NULL, (void *)&fileP, &fileSize);
  if (rc != 0)
  {
    code = 3;
    goto out;
  }

  /* fsnotify_open(fileP->f_dentry); */
  fd_install(fd, fileP);
  *fdP = fd;

out:
  if (vfsP != NULL)
    (void)cxiCacheExit(vfsP, NULL); /* ignore error */

  TRACE8(TRACE_PCACHE, 4, TRCID_CACHE_NFS_OPEN,
         "cxiCacheNfsOpen: sgname %s filesetId %d parent ino %d name %s flags %d "
         "fd %d code %d rc %d",
         sgNameP, filesetId, pattrP->pa_vattr.va_ino, nameP, flags, fd, code, rc);
  return rc;
}

int
cxiCacheGetFilename(void *vfsP, unsigned long pino, unsigned long cino,
                    char *nameP, int foldVal)
{
  int rc = ENOENT, code = 0;
  struct super_block *sb = (struct super_block *)vfsP;
  struct dentry *pDentry = NULL, *cDentry = NULL;
  unsigned int genNum = 0xffffffff;

  pDentry = get_dentry(sb, pino, genNum);
  if (pDentry == NULL || IS_ERR(pDentry))
  {
    TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_GET_FNAME_1,
           "cxiCacheGetFilename pDentry 0x%X err %d",
            pDentry, PTR_ERR(pDentry));
    code = 1;
    goto out;
  }
  cDentry = get_dentry(sb, cino, genNum);
  if (cDentry == NULL || IS_ERR(cDentry))
  {
    TRACE2(TRACE_PCACHE, 4, TRCID_CACHE_GET_FNAME_2,
           "cxiCacheGetFilename cDentry 0x%X err %d",
           cDentry, PTR_ERR(cDentry));
    dput(pDentry);
    code = 2;
    goto out;
  }

  rc = gpfs_get_name_by_fold(pDentry, nameP, cDentry, foldVal);
  dput(pDentry);
  dput(cDentry);

out:
  TRACE6(TRACE_PCACHE, 4, TRCID_CACHE_GET_FNAME,
         "cxiCacheGetFilename: sb 0x%X pino %d cino %d name %s code %d rc %d",
         sb, pino, cino, nameP, code, rc);
  return rc;
}
